from typing import List, Optional

from pydantic import BaseModel

from phi.enums.response_enums import BackendResponseStatus


class BackendResponse(BaseModel):
    status: BackendResponseStatus = BackendResponseStatus.FAIL
    message: str = "INVALID"
    message_log: Optional[List[str]] = None



import datetime
from typing import Any, Dict, Optional

from pydantic import BaseModel

from phi.utils.dttm import dttm_str_to_dttm
from phi.workspace.ws_enums import WorkspaceVisibilityEnum

######################################################
# WorkspaceSchema Schemas
# These should match ~/philab/services/zeus/zeus/schemas/workspaces.py
######################################################


# The WorkspaceSchema defines the properties which are returned by the backend api
# Need to be careful that this doesn't contain any sensitive info
class WorkspaceSchema(BaseModel):
    name: str
    visibility: WorkspaceVisibilityEnum
    git_url: Optional[str] = None
    is_active: bool
    is_verified: bool
    is_test: bool
    setup_complete: bool
    id_workspace: int
    created_by_id_user: int
    created_by_email: str
    create_ts: datetime.datetime
    last_update_ts: Optional[datetime.datetime] = None
    extra_data: Optional[Dict[str, Any]] = None

    @classmethod
    def from_dict(cls, workspace_dict: Dict[str, Any]):
        return cls(
            name=workspace_dict.get("name", None),
            id_workspace=workspace_dict.get("id_workspace", None),
            visibility=WorkspaceVisibilityEnum.from_str(
                workspace_dict.get("visibility", None)
            ),
            git_url=workspace_dict.get("git_url", None),
            is_active=workspace_dict.get("is_valid", False),
            is_verified=workspace_dict.get("is_verified", False),
            is_test=workspace_dict.get("is_test", False),
            setup_complete=workspace_dict.get("setup_complete", False),
            created_by_id_user=workspace_dict.get("created_by_id_user", None),
            created_by_email=workspace_dict.get("created_by_email", None),
            create_ts=dttm_str_to_dttm(workspace_dict.get("create_ts", None)),
            last_update_ts=dttm_str_to_dttm(workspace_dict.get("last_update_ts", None)),
            extra_data=workspace_dict.get("extra_data", None),
        )


class UpsertWorkspaceFromCli(BaseModel):
    name: str
    visibility: Optional[str] = None
    git_url: Optional[str] = None
    setup_complete: Optional[bool] = None
    extra_data: Optional[Dict[str, Any]] = None
    is_active_ws_for_user: Optional[bool] = None
    # The UpsertWorkspaceFromCli model can be used to create or update a workspace
    # But sometimes we want to use it for ONLY creation or ONLY update
    # These 2 flags help us limit the functionality
    # If true, we can ONLY CREATE a workspace, not UPDATE it
    only_create: bool = False
    # If true, we can ONLY UPDATE a workspace, not CREATE it
    only_update: bool = False


class UpdateWorkspace(BaseModel):
    id_workspace: int
    name: Optional[str]
    visibility: Optional[WorkspaceVisibilityEnum]
    git_url: Optional[str]
    is_active: Optional[bool]
    is_verified: Optional[bool]
    is_test: Optional[bool]
    setup_complete: Optional[bool]
    is_primary_ws_for_user: Optional[bool]
    extra_data: Optional[Dict[str, Any]] = None



from pathlib import Path


def rmdir_recursive(dir: Path) -> bool:
    """Deletes a dir recursively, including all files and dirs in that directory"""

    if not dir.exists():
        return True

    for item in dir.iterdir():
        if item.is_dir():
            rmdir_recursive(item)
        else:
            item.unlink()
    dir.rmdir()
    return True if not dir.exists() else False


def delete_files_in_dir(dir: Path) -> None:
    """Deletes all files in a directory, but doesn't delete the directory"""

    for item in dir.iterdir():
        if item.is_dir():
            rmdir_recursive(item)
        else:
            item.unlink()


def delete_from_fs(path_to_del: Path) -> bool:
    if not path_to_del.exists():
        return True
    if path_to_del.is_dir():
        return rmdir_recursive(path_to_del)
    else:
        path_to_del.unlink()
    return True if not path_to_del.exists() else False



from collections import OrderedDict
from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple

from phi.conf.constants import PHI_CONF_PATH
from phi.schemas.user_schemas import UserSchema
from phi.utils.cli_console import (
    print_available_workspaces,
    print_info,
    print_error,
    print_heading,
    print_horizontal_line,
    print_info,
    print_subheading,
)
from phi.utils.common import pickle_object_to_file, unpickle_object_from_file
from phi.utils.dttm import current_datetime_utc
from phi.utils.filesystem import delete_from_fs
from phi.utils.log import logger
from phi.workspace.phi_ws_data import PhiWsData
from phi.workspace.ws_enums import WorkspaceSetupActions
from phi.workspace.ws_schemas import WorkspaceSchema


class PhiConf:
    """The PhiConf class manages user data for the phi cli"""

    def __init__(self):
        # Current user, populated after authenticating with the backend
        # To add a user, use the user setter
        self._user: Optional[UserSchema] = None

        # Active ws - used as the default for all phi commands
        # To add an active workspace, use the active_ws_name setter
        self._active_ws_name: Optional[str] = None

        # ** The self._ws_data_map holds the mapping from ws_name to ws_data.
        self._ws_data_map: Dict[str, PhiWsData] = OrderedDict()
        # When phi is initialized using `phi init`, this is an empty dict
        # When the user is authenticated, we get a list of all the workspaces this user has access to.
        # These workspaces are then added to PhiConf using the `active_ws_name` and `available_ws` setters.
        # Note: These workspaces arent yet cloned on the users machines and only a record of their availability is kept.
        # When the workspace is setup on the users machine, we update the remaining workspace data.

        # Quick access from ws_path -> ws data.
        self._path_to_ws_data_map: Dict[Path, PhiWsData] = OrderedDict()

        # Save the config to disk once initialized
        self.save_config()

    ######################################################
    ## User functions
    ######################################################

    @property
    def user(self) -> Optional[UserSchema]:
        return self._user

    @user.setter
    def user(self, user: Optional[UserSchema]) -> None:
        """Sets the user"""
        if user is not None:
            self._user = user
            self.save_config()

    ######################################################
    ## Workspace functions
    ######################################################

    @property
    def active_ws_name(self) -> Optional[str]:
        return self._active_ws_name

    @active_ws_name.setter
    def active_ws_name(self, ws_name: Optional[str]) -> None:
        if ws_name is not None:
            self._active_ws_name = ws_name
            self._add_or_update_ws_data(ws_name=ws_name)
            self.save_config()

    @property
    def available_ws(self) -> List[WorkspaceSchema]:
        return [
            w.ws_schema for w in self._ws_data_map.values() if w.ws_schema is not None
        ]

    @available_ws.setter
    def available_ws(self, avl_ws: Optional[List[WorkspaceSchema]]) -> None:
        if avl_ws:
            for ws_schema in avl_ws:
                self._add_or_update_ws_data(ws_name=ws_schema.name, ws_schema=ws_schema)
            self.save_config()

    def _add_or_update_ws_data(
        self,
        ws_name: str,
        ws_schema: Optional[WorkspaceSchema] = None,
        ws_dir_path: Optional[Path] = None,
        ws_config_file_path: Optional[Path] = None,
    ) -> None:
        """The main function to create, update or refresh PhiWsData in PhiConf.

        Notes:
        * ws_name is the only required argument.
        * This function does not call self.save_config(). call it from the parent.

        This function is called from 5 places:
        1. During `phi auth` when active/available workspaces are requested from the backend.
        2. During `phi ws init` when a new workspace is created on the users machine.
        3. When the user already has a local workspace dir on their machine
            and need to manually add the ws to the PhiConf
            (possible after running `phi init -r` which deletes the config)
        4. Whenever ws config, creds or data needs to be refreshed. This function is called with just the `ws_name` arg
        5. When active or available workspaces are updated, this function is called from the
            @active_ws_name.setter and @available_ws.setter with the ws_name and ws_schema
        """

        ######################################################
        # Validate ws_name
        ######################################################

        # ws_name should not be None
        if ws_name is None or not isinstance(ws_name, str):
            return

        ######################################################
        # Create ws_data if it does not exist
        ######################################################

        if ws_name not in self._ws_data_map:
            logger.debug(f"Creating workspace: {ws_name}")
            self._ws_data_map[ws_name] = PhiWsData(
                ws_name=ws_name,
                ws_schema=ws_schema,
                ws_dir_path=ws_dir_path,
                ws_config_file_path=ws_config_file_path,
            )
            if ws_dir_path is not None:
                self._path_to_ws_data_map[ws_dir_path] = self._ws_data_map[ws_name]
                logger.debug(f"Workspace dir: {ws_dir_path}")
            logger.debug(f"Workspace created: `{ws_name}`")

        ######################################################
        # Update ws_data
        ######################################################

        # By this point there should be a PhiWsData object for this ws_name
        existing_ws_data: PhiWsData = self._ws_data_map.get(
            ws_name, PhiWsData(ws_name=ws_name)
        )
        # Make a deep copy of the existing workspace data
        # This allows us to update the fields we want, and keep any existing fields as is
        _ws_data: PhiWsData = existing_ws_data.copy(deep=True)

        # Update values where available
        if ws_schema is not None and ws_schema != existing_ws_data.ws_schema:
            logger.debug("Updating ws_schema")
            _ws_data.ws_schema = ws_schema
        if ws_dir_path is not None and ws_dir_path != existing_ws_data.ws_dir_path:
            logger.debug("Updating ws_dir_path")
            _ws_data.ws_dir_path = ws_dir_path
        if (
            ws_config_file_path is not None
            and ws_config_file_path != existing_ws_data.ws_config_file_path
        ):
            logger.debug("Updating ws_config_file_path")
            _ws_data.ws_config_file_path = ws_config_file_path

        # Load the WorkspaceConfig
        _ws_data.load_workspace_from_config_file()
        _ws_data.last_update_ts = current_datetime_utc()

        # Point the ws_data in _ws_data_map and _path_to_ws_data_map
        # to new _ws_data
        logger.debug(f"Updating mapping for: {ws_name}")

        # 1. Pop the existing object from the self._ws_data_map
        if ws_name in self._ws_data_map:
            popped_ws_data_from_map: Optional[PhiWsData] = self._ws_data_map.pop(
                ws_name
            )
            # logger.debug(f"Removed {ws_name} from _ws_data_map")
        self._ws_data_map[ws_name] = _ws_data

        # 2. Pop the existing object from the self._path_to_ws_data_map
        if (
            _ws_data.ws_dir_path is not None
            and _ws_data.ws_dir_path in self._path_to_ws_data_map
        ):
            popped_ws_data_from_path_map: Optional[
                PhiWsData
            ] = self._path_to_ws_data_map.pop(existing_ws_data.ws_dir_path)
            # logger.debug(f"Removed {ws_name} from _path_to_ws_data_map")
        self._path_to_ws_data_map[_ws_data.ws_dir_path] = self._ws_data_map[ws_name]

        ######################################################
        # END
        ######################################################

    def add_new_created_ws_to_config(
        self,
        ws_name: str,
        ws_dir_path: Path,
        ws_config_file_path: Path,
    ) -> None:
        """Adds a newly created workspace to the PhiConf"""
        self._add_or_update_ws_data(
            ws_name=ws_name,
            ws_dir_path=ws_dir_path,
            ws_config_file_path=ws_config_file_path,
        )
        self.save_config()

    def update_ws_schema_for_workspace(
        self,
        ws_schema: WorkspaceSchema,
        set_as_active: bool = False,
    ) -> None:
        """Updates the schema for a workspace and sets as active if needed"""
        self._add_or_update_ws_data(
            ws_name=ws_schema.name,
            ws_schema=ws_schema,
        )
        if set_as_active:
            self.active_ws_name = ws_schema.name
        self.save_config()

    def update_ws_data(
        self,
        ws_name: str,
        ws_schema: Optional[WorkspaceSchema] = None,
        ws_dir_path: Optional[Path] = None,
        ws_config_file_path: Optional[Path] = None,
    ) -> None:
        """Update workspace data and return True if successful"""

        logger.debug(f"Updating ws_data for {ws_name}")
        self._add_or_update_ws_data(
            ws_name=ws_name,
            ws_schema=ws_schema,
            ws_dir_path=ws_dir_path,
            ws_config_file_path=ws_config_file_path,
        )
        self.save_config()

    def update_ws_data_for_cloned_ws(
        self,
        ws_name: str,
        ws_dir_path: Path,
        ws_config_file_path: Path,
    ) -> bool:
        """For a workspace to be cloned, it would already have ws_data and ws_schema available.
        After the directory from the git_url is cloned, we can fill in the gaps to update
        the ws_data for the workspace. Call this function through ws_operator.clone_workspace()
        """
        ws_data: Optional[PhiWsData] = self._ws_data_map.get(ws_name, None)
        if ws_data is None:
            return False

        self._add_or_update_ws_data(
            ws_name=ws_name,
            ws_dir_path=ws_dir_path,
            ws_config_file_path=ws_config_file_path,
        )
        self.save_config()
        return True

    def add_ws_data_using_config_file(
        self, ws_dir_path: Path, ws_config_file_path: Optional[Path]
    ) -> Optional[PhiWsData]:
        """Maps a ws at `ws_dir_path` to an available workspace. using the config file at `ws_config_file_path`.
        Also updates the ws_config_file_path in the ws_data if provided.

        This is used in 2 cases:
        1. If the user ran `phi init -r`, the PhiConf gets erased and so does self._path_to_ws_data_map
            So there is no link from the ws_dir_path to the workspace.
        2. The user manually cloned the workspace directory. In that case we don't have a record of this
            ws_dir_path matching any available workspace.

        This function basically reads the workspace config and tries to map it to an available workspace.
        Returns the PhiWsData if successful
        """

        logger.debug(f"Looking for the workspace at {ws_dir_path}")
        _best_guess_ws_name = ws_dir_path.stem
        logger.debug(f"_best_guess_ws_name: {_best_guess_ws_name}")
        _ws_schema: Optional[WorkspaceSchema] = None
        for avl_ws in self.available_ws:
            if avl_ws.name == _best_guess_ws_name:
                _ws_schema = avl_ws
            break
        if _ws_schema is None:
            print_error(
                f"Workspace name: {_best_guess_ws_name} does not match any available workspaces"
            )
            print_available_workspaces(self.available_ws)
            return None

        logger.debug(f"Found matching workspace {_ws_schema.name}")
        logger.debug(
            f"Mapping {_ws_schema.name} to dir: {ws_dir_path} and config: {ws_config_file_path}"
        )
        self._add_or_update_ws_data(
            ws_name=_ws_schema.name,
            ws_schema=_ws_schema,
            ws_dir_path=ws_dir_path,
            ws_config_file_path=ws_config_file_path,
        )
        self.save_config()
        return self._ws_data_map.get(_ws_schema.name, None)

    def validate_workspace(
        self, ws_name: str
    ) -> Tuple[bool, Optional[Set[WorkspaceSetupActions]]]:
        """Validates a workspace and returns a tuple of [validation_status, pending_setup_actions]
        Returns:
            (False, None): If PhiWsData is not valid
            (True, None): If Validation is successful and everything is properly setup
            (True, set(WorkspaceSetupActions)): If Validation is successful but user has pending setup actions
            (False, set(WorkspaceSetupActions)): If Validation is unsuccessful and user has pending setup actions
        """

        # Validate the ws has a valid entry
        if ws_name is None or ws_name not in self._ws_data_map:
            return False, None
        logger.debug(f"Validating workspace: {ws_name}")

        ws_data: PhiWsData = self._ws_data_map[ws_name]

        # If ws_config is missing add a pending action
        if ws_data.ws_config is None:
            ws_data.pending_actions.add(WorkspaceSetupActions.WS_CONFIG_IS_AVL)
        else:
            ws_data.pending_actions.discard(WorkspaceSetupActions.WS_CONFIG_IS_AVL)

        # If ws_schema is missing add a pending action
        if ws_data.ws_schema is None:
            ws_data.pending_actions.add(WorkspaceSetupActions.WS_IS_AUTHENTICATED)
        else:
            ws_data.pending_actions.discard(WorkspaceSetupActions.WS_IS_AUTHENTICATED)

        # If git_url is missing add a pending action
        if ws_data.ws_schema is None:
            from phi.utils.git import get_remote_origin_for_dir

            _remote_origin = get_remote_origin_for_dir(ws_data.ws_dir_path)
            if _remote_origin is None:
                ws_data.pending_actions.add(
                    WorkspaceSetupActions.GIT_REMOTE_ORIGIN_IS_AVL
                )
        elif ws_data.ws_schema.git_url is None:
            ws_data.pending_actions.add(WorkspaceSetupActions.GIT_REMOTE_ORIGIN_IS_AVL)
        else:
            ws_data.pending_actions.discard(
                WorkspaceSetupActions.GIT_REMOTE_ORIGIN_IS_AVL
            )

        # logger.debug(f"Required Actions for this workspace: {ws_data.required_actions}")
        # logger.debug(f"Pending Actions for this workspace: {ws_data.pending_actions}")

        # Check how many required_actions are pending
        if len(ws_data.required_actions.intersection(ws_data.pending_actions)) > 0:
            return False, ws_data.pending_actions

        return True, ws_data.pending_actions

    def delete_ws(self, ws_name: str, delete_ws_dir: bool = False) -> bool:
        """Handles Deleting a workspace from the PhiConf"""

        print_info(f"Deleting workspace: {ws_name}")
        print_info(f"--> this does not delete any files, please delete them manually")

        ws_data: Optional[PhiWsData] = self._ws_data_map.pop(ws_name, None)
        if ws_data is None:
            print_error(f"No record of workspace {ws_name}")
            return False

        ws_data_from_path: Optional[PhiWsData]
        if ws_data.ws_dir_path is not None:
            ws_data_from_path = self._path_to_ws_data_map.pop(ws_data.ws_dir_path, None)

        # Check if we're deleting the active workspace, if yes, unset the active ws
        if (
            self._active_ws_name is not None
            and ws_data.ws_name is not None
            and self._active_ws_name == ws_data.ws_name
        ):
            print_info(f"Removing {ws_data.ws_name} as the active workspace")
            self._active_ws_name = None

        ws_dir_path_deleted: bool = False
        if delete_ws_dir and ws_data.ws_dir_path:
            print_info(f"Deleting {str(ws_data.ws_dir_path)}")
            ws_dir_path_deleted = delete_from_fs(ws_data.ws_dir_path)

        # Save the config after deleting workspace
        self.save_config()
        return True

    ######################################################
    ## Get Workspace Data
    ######################################################

    def get_ws_data_by_name(self, ws_name: str) -> Optional[PhiWsData]:
        return self._ws_data_map[ws_name] if ws_name in self._ws_data_map else None

    def get_ws_data_by_path(self, ws_dir_path: Path) -> Optional[PhiWsData]:
        return (
            self._path_to_ws_data_map[ws_dir_path]
            if ws_dir_path in self._path_to_ws_data_map
            else None
        )

    ## Workspace Name
    def get_ws_name_by_path(self, ws_dir_path: Path) -> Optional[str]:
        if ws_dir_path in self._path_to_ws_data_map:
            return self._path_to_ws_data_map[ws_dir_path].ws_name
        return None

    ## Workspace Schema
    def get_ws_schema_by_name(self, ws_name: str) -> Optional[WorkspaceSchema]:
        if ws_name in self._ws_data_map:
            return self._ws_data_map[ws_name].ws_schema
        return None

    def get_ws_schema_by_path(self, ws_dir_path: Path) -> Optional[WorkspaceSchema]:
        if ws_dir_path in self._path_to_ws_data_map:
            return self._path_to_ws_data_map[ws_dir_path].ws_schema
        return None

    ## Workspace Directory
    def get_ws_dir_path_by_name(self, ws_name: str) -> Optional[Path]:
        if ws_name in self._ws_data_map:
            return self._ws_data_map[ws_name].ws_dir_path
        return None

    ## Workspace Config File Path
    def get_ws_config_file_path_by_name(self, ws_name: str) -> Optional[Path]:
        if ws_name in self._ws_data_map:
            return self._ws_data_map[ws_name].ws_config_file_path
        return None

    ## Workspace Pending Actions
    def get_ws_pending_actions_by_name(
        self, ws_name: str
    ) -> Optional[Set[WorkspaceSetupActions]]:
        if ws_name in self._ws_data_map:
            return self._ws_data_map[ws_name].pending_actions
        return None

    ## Workspace Data
    def get_active_ws_data(self, refresh: bool = False) -> Optional[PhiWsData]:
        if self.active_ws_name is not None and self.active_ws_name in self._ws_data_map:
            if refresh:
                self.refresh_ws_config(self.active_ws_name)
            return self._ws_data_map[self.active_ws_name]
        return None

    def refresh_ws_config(self, ws_name: str) -> None:
        """Refresh the workspace config for a given workspace name"""

        self._add_or_update_ws_data(ws_name=ws_name)
        logger.debug(f"++**++ Config refreshed: {ws_name}")
        self.save_config()

    ######################################################
    ## PhiConf functions
    ######################################################

    def save_config(self):
        # logger.debug(f"Saving config to {str(PHI_CONF_PATH)}")
        pickle_object_to_file(self, PHI_CONF_PATH)

    @classmethod
    def get_saved_conf(cls):
        # logger.debug(f"Reading phidata conf at {PHI_CONF_PATH}")
        if PHI_CONF_PATH.exists():
            # logger.debug(f"{PHI_CONF_PATH} exists")
            if PHI_CONF_PATH.is_file():
                return unpickle_object_from_file(
                    file_path=PHI_CONF_PATH, verify_class=cls
                )
            elif PHI_CONF_PATH.is_dir():
                logger.debug(f"{PHI_CONF_PATH} is a directory, deleting and exiting")
                delete_from_fs(PHI_CONF_PATH)
        return None

    def print_to_cli(self):
        print_heading("Phidata Config\n")
        if self.user:
            print_info(f"* User: {self.user.email}")
        if self.active_ws_name:
            print_info(f"* Active workspace: {self.active_ws_name}")
        else:
            print_info("* No active workspace, run `phi ws init` to create one")
        if len(self._ws_data_map) > 0:
            print_subheading("\nAvailable workspaces")
            for k, v in self._ws_data_map.items():
                print_info(f"Name: {v.ws_name}")
                print_info(f"Dir: {v.ws_dir_path}\n")
                print_horizontal_line()



from typing import List, Optional, cast, Dict, Any
from pathlib import Path

from docker.models.containers import Container
from phidata.infra.docker.api_client import DockerApiClient
from phidata.infra.docker.config import DockerConfig
from phidata.infra.docker.exceptions import DockerConfigException
from phidata.infra.docker.manager import DockerManager
from phidata.infra.docker.utils.container import execute_command
from phidata.product import DataProduct
from phidata.workflow import Workflow
from phidata.types.context import PathContext, RunContext

from phi.types.run_status import RunStatus
from phi.utils.cli_console import (
    print_error,
    print_heading,
    print_info,
    print_warning,
    print_subheading,
)
from phi.utils.log import logger


def deploy_docker_config(
    config: DockerConfig,
    name_filter: Optional[str] = None,
    type_filter: Optional[str] = None,
    dry_run: Optional[bool] = False,
) -> bool:

    # Step 1: Get the DockerManager
    docker_manager: DockerManager = config.get_docker_manager()
    if docker_manager is None:
        raise DockerConfigException("DockerManager unavailable")

    # Step 2: If dry_run, print the resources and return True
    if dry_run:
        docker_manager.create_resources_dry_run(
            name_filter=name_filter, type_filter=type_filter
        )
        return True

    # Step 3: Create resources
    env = config.env
    print_heading(
        "Deploying docker config{}\n".format(
            f" for env: {env}" if env is not None else ""
        )
    )
    try:
        success: bool = docker_manager.create_resources(
            name_filter=name_filter, type_filter=type_filter
        )
        if not success:
            return False
    except Exception:
        raise

    # Step 4: Validate resources are created
    resource_creation_valid: bool = docker_manager.validate_resources_are_created(
        name_filter=name_filter, type_filter=type_filter
    )
    if not resource_creation_valid:
        logger.error("DockerResource creation could not be validated")
        return False

    print_info("Docker config deployed")
    return True


def shutdown_docker_config(
    config: DockerConfig,
    name_filter: Optional[str] = None,
    type_filter: Optional[str] = None,
    dry_run: Optional[bool] = False,
) -> bool:

    # Step 1: Get the DockerManager
    docker_manager: DockerManager = config.get_docker_manager()
    if docker_manager is None:
        raise DockerConfigException("DockerManager unavailable")

    # Step 2: If dry_run, print the resources and return True
    if dry_run:
        docker_manager.delete_resources_dry_run(
            name_filter=name_filter, type_filter=type_filter
        )
        return True

    # Step 3: Delete resources
    env = config.env
    print_heading(
        "Shutting down docker config{}\n".format(
            f" for env: {env}" if env is not None else ""
        )
    )
    try:
        success: bool = docker_manager.delete_resources(
            name_filter=name_filter, type_filter=type_filter
        )
        if not success:
            return False
    except Exception:
        raise

    # Step 4: Validate resources are delete
    resources_deletion_valid: bool = docker_manager.validate_resources_are_deleted(
        name_filter=name_filter, type_filter=type_filter
    )
    if not resources_deletion_valid:
        logger.error("DockerResource deletion could not be validated")
        return False

    print_info("Workspace shutdown")
    return True


def run_workflows_docker(
    workflow_file: str,
    workflows: Dict[str, Workflow],
    run_context: RunContext,
    docker_config: DockerConfig,
    target_app: Optional[str] = None,
    workflow_name: Optional[str] = None,
    use_dag_id: Optional[str] = None,
) -> bool:
    from phidata.app.devbox import DevboxArgs, default_devbox_name
    from phidata.infra.docker.resource.types import (
        DockerResourceType,
        DockerContainer,
    )

    logger.debug("Running Workflow in DockerContainer")
    # Step 1: Get the DockerManager
    docker_manager: DockerManager = docker_config.get_docker_manager()
    if docker_manager is None:
        raise DockerConfigException("DockerManager unavailable")

    # Step 2: Check if a Devbox is available for running the Workflow
    # If available get the DevboxArgs
    devbox_app_name = target_app or docker_config.devbox_name or default_devbox_name
    logger.debug(f"Using App: {devbox_app_name}")
    devbox_app = docker_config.get_app_by_name(devbox_app_name)
    if devbox_app is None:
        print_error("Devbox not available")
        return False
    devbox_app_args: Optional[Any] = devbox_app.args
    # logger.debug(f"DevboxArgs: {devbox_app_args}")
    if devbox_app_args is None or not isinstance(devbox_app_args, DevboxArgs):
        print_error("DevboxArgs invalid")
        return False
    devbox_app_args = cast(DevboxArgs, devbox_app_args)

    # Step 3: Build the PathContext for the workflows.
    # NOTE: The PathContext uses directories relative to the
    # workspace_parent_container_path
    workspace_name = docker_config.workspace_root_path.stem
    workspace_root_container_path = Path(
        devbox_app_args.workspace_parent_container_path
    ).joinpath(workspace_name)
    scripts_dir_container_path = workspace_root_container_path.joinpath(
        docker_config.scripts_dir
    )
    storage_dir_container_path = workspace_root_container_path.joinpath(
        docker_config.storage_dir
    )
    meta_dir_container_path = workspace_root_container_path.joinpath(
        docker_config.meta_dir
    )
    products_dir_container_path = workspace_root_container_path.joinpath(
        docker_config.products_dir
    )
    notebooks_dir_container_path = workspace_root_container_path.joinpath(
        docker_config.notebooks_dir
    )
    workspace_config_dir_container_path = workspace_root_container_path.joinpath(
        docker_config.workspace_config_dir
    )
    workflow_file_path = Path(products_dir_container_path).joinpath(workflow_file)
    wf_path_context: PathContext = PathContext(
        scripts_dir=scripts_dir_container_path,
        storage_dir=storage_dir_container_path,
        meta_dir=meta_dir_container_path,
        products_dir=products_dir_container_path,
        notebooks_dir=notebooks_dir_container_path,
        workspace_config_dir=workspace_config_dir_container_path,
        workflow_file=workflow_file_path,
    )
    logger.debug(f"PathContext: {wf_path_context.json(indent=4)}")

    # Step 4: Get the container to run the Workflow
    devbox_containers: Optional[
        List[DockerResourceType]
    ] = docker_manager.get_resources(
        name_filter=devbox_app_name, type_filter="DockerContainer"
    )
    # logger.debug(f"devbox_containers: {devbox_containers}")
    if devbox_containers is None:
        logger.error(f"DockerContainer:{devbox_app_name} not found")
        return False
    if len(devbox_containers) > 1:
        print_warning(
            "Running commands in multiple containers is not yet supported. "
            + "Running in the first container. "
        )
    devbox_container: DockerContainer = devbox_containers[0]
    # logger.debug("devbox_container: ")
    # logger.debug("Name: {}".format(devbox_container.name))
    # logger.debug("Class: {}".format(devbox_container.__class__))
    # logger.debug("Resource: {}".format(devbox_container))
    docker_client: DockerApiClient = docker_manager.docker_worker.docker_client
    active_container: Optional[Container] = devbox_container.read(docker_client)
    # logger.debug("active_container: {}".format(active_container.attrs))
    # logger.debug("Class: {}".format(active_container.__class__))
    # logger.debug("Type: {}".format(type(active_container)))
    if active_container is None or not isinstance(active_container, Container):
        print_error("Container not available, please check your workspace is deployed.")
        return False

    # Step 5: Run single Workflow if workflow_name is provided
    if workflow_name is not None:
        if workflow_name not in workflows:
            print_error(
                "Could not find '{}' in {}".format(
                    workflow_name, "[{}]".format(", ".join(workflows.keys()))
                )
            )
            return False

        wf_run_success: bool = False
        workflow_to_run = workflows[workflow_name]
        _name = workflow_name or workflow_to_run.name
        print_subheading(f"\nRunning {_name}")
        # Pass down context
        workflow_to_run.run_context = run_context
        workflow_to_run.path_context = wf_path_context
        # Use DataProduct dag_id if provided
        if use_dag_id is not None:
            workflow_to_run.dag_id = use_dag_id
        wf_run_success = workflow_to_run.run_in_docker_container(
            active_container=active_container, docker_env=docker_config.docker_env
        )

        print_subheading("\nWorkflow run status:")
        print_info("{}: {}".format(_name, "Success" if wf_run_success else "Fail"))
        print_info("")
        return wf_run_success
    # Step 6: Run all Workflows if workflow_name is None
    else:
        wf_run_status: List[RunStatus] = []
        for wf_name, wf_obj in workflows.items():
            _name = wf_name or wf_obj.name
            print_subheading(f"\nRunning {_name}")
            # Pass down context
            wf_obj.run_context = run_context
            wf_obj.path_context = wf_path_context
            run_success = wf_obj.run_in_docker_container(
                active_container=active_container, docker_env=docker_config.docker_env
            )
            wf_run_status.append(RunStatus(_name, run_success))

        print_subheading("\nWorkflow run status:")
        print_info(
            "\n".join(
                [
                    "{}: {}".format(wf.name, "Success" if wf.success else "Fail")
                    for wf in wf_run_status
                ]
            )
        )
        print_info("")
        for _run in wf_run_status:
            if not _run.success:
                return False
        return True


def run_data_products_docker(
    workflow_file: str,
    data_products: Dict[str, DataProduct],
    run_context: RunContext,
    docker_config: DockerConfig,
    target_app: Optional[str] = None,
) -> bool:
    from phidata.app.devbox import DevboxArgs, default_devbox_name
    from phidata.infra.docker.resource.types import (
        DockerResourceType,
        DockerContainer,
    )

    logger.debug("Running DataProducts in DockerContainer")
    # Step 1: Get the DockerManager
    docker_manager: DockerManager = docker_config.get_docker_manager()
    if docker_manager is None:
        raise DockerConfigException("DockerManager unavailable")

    # Step 2: Check if a DevboxApp is available for running the DataProduct
    # If available get the DevboxAppArgs
    devbox_app_name = target_app or docker_config.devbox_name or default_devbox_name
    logger.debug(f"Using App: {devbox_app_name}")
    devbox_app = docker_config.get_app_by_name(devbox_app_name)
    if devbox_app is None:
        print_error("Devbox not available")
        return False
    devbox_app_args: Optional[Any] = devbox_app.args
    # logger.debug(f"DevboxAppArgs: {devbox_app_args}")
    if devbox_app_args is None or not isinstance(devbox_app_args, DevboxArgs):
        print_error("DevboxArgs invalid")
        return False
    devbox_app_args = cast(DevboxArgs, devbox_app_args)

    # Step 3: Build the PathContext for the DataProducts.
    # NOTE: The PathContext uses directories relative to the workspace_parent_container_path
    workspace_name = docker_config.workspace_root_path.stem
    workspace_root_container_path = Path(
        devbox_app_args.workspace_parent_container_path
    ).joinpath(workspace_name)
    scripts_dir_container_path = workspace_root_container_path.joinpath(
        docker_config.scripts_dir
    )
    storage_dir_container_path = workspace_root_container_path.joinpath(
        docker_config.storage_dir
    )
    meta_dir_container_path = workspace_root_container_path.joinpath(
        docker_config.meta_dir
    )
    products_dir_container_path = workspace_root_container_path.joinpath(
        docker_config.products_dir
    )
    notebooks_dir_container_path = workspace_root_container_path.joinpath(
        docker_config.notebooks_dir
    )
    workspace_config_dir_container_path = workspace_root_container_path.joinpath(
        docker_config.workspace_config_dir
    )
    workflow_file_path = Path(products_dir_container_path).joinpath(workflow_file)
    dp_path_context: PathContext = PathContext(
        scripts_dir=scripts_dir_container_path,
        storage_dir=storage_dir_container_path,
        meta_dir=meta_dir_container_path,
        products_dir=products_dir_container_path,
        notebooks_dir=notebooks_dir_container_path,
        workspace_config_dir=workspace_config_dir_container_path,
        workflow_file=workflow_file_path,
    )
    logger.debug(f"PathContext: {dp_path_context.json(indent=4)}")

    # Step 4: Get the container to run the DataProduct
    devbox_containers: Optional[
        List[DockerResourceType]
    ] = docker_manager.get_resources(
        name_filter=devbox_app_name, type_filter="DockerContainer"
    )
    if devbox_containers is None:
        logger.error(f"DockerContainer:{devbox_app_name} not found")
        return False
    if len(devbox_containers) > 1:
        print_info(
            "Running commands in multiple containers is not yet supported. "
            + "Running in the first container. "
        )
    devbox_container: DockerContainer = devbox_containers[0]
    # logger.debug("devbox_container: ")
    # logger.debug("Name: {}".format(devbox_container.name))
    # logger.debug("Class: {}".format(devbox_container.__class__))
    # logger.debug("Resource: {}".format(devbox_container))
    docker_client: DockerApiClient = docker_manager.docker_worker.docker_client
    active_container: Optional[Container] = devbox_container.read(docker_client)
    # logger.debug("active_container: {}".format(active_container.attrs))
    # logger.debug("Class: {}".format(active_container.__class__))
    # logger.debug("Type: {}".format(type(active_container)))
    if active_container is None or not isinstance(active_container, Container):
        print_error("Container not available, please check your workspace is deployed.")
        return False

    # Step 5: Run the DataProducts
    dp_run_status: List[RunStatus] = []
    for dp_name, dp_obj in data_products.items():
        _name = dp_name or dp_obj.name
        print_subheading(f"\nRunning {_name}")
        # Pass down context
        dp_obj.run_context = run_context
        dp_obj.path_context = dp_path_context
        run_success = dp_obj.run_in_docker_container(
            active_container=active_container, docker_env=docker_config.docker_env
        )
        dp_run_status.append(RunStatus(_name, run_success))

    print_subheading("DataProduct run status:")
    print_info(
        "\n".join(
            [
                "{}: {}".format(wf.name, "Success" if wf.success else "Fail")
                for wf in dp_run_status
            ]
        )
    )
    print_info("")
    for _run in dp_run_status:
        if not _run.success:
            return False
    return True


def run_command_docker(
    command: str,
    docker_config: DockerConfig,
    target_app: Optional[str] = None,
) -> bool:
    from phidata.app.devbox import DevboxArgs, default_devbox_name
    from phidata.infra.docker.resource.types import (
        DockerResourceType,
        DockerContainer,
    )

    logger.debug("Running command in DockerContainer")
    # Step 1: Get the DockerManager
    docker_manager: DockerManager = docker_config.get_docker_manager()
    if docker_manager is None:
        raise DockerConfigException("DockerManager unavailable")

    # Step 2: Check if a Devbox is available for running the DataProduct
    # If available get the DevboxArgs
    devbox_app_name = target_app or docker_config.devbox_name or default_devbox_name
    logger.debug(f"Using App: {devbox_app_name}")
    devbox_app = docker_config.get_app_by_name(devbox_app_name)
    if devbox_app is None:
        print_error("Devbox not available")
        return False
    devbox_app_args: Optional[Any] = devbox_app.args
    # logger.debug(f"DevboxArgs: {devbox_app_args}")
    if devbox_app_args is None or not isinstance(devbox_app_args, DevboxArgs):
        print_error("DevboxArgs invalid")
        return False
    devbox_app_args = cast(DevboxArgs, devbox_app_args)

    # Step 3: Get the container to run the command
    devbox_containers: Optional[
        List[DockerResourceType]
    ] = docker_manager.get_resources(
        name_filter=devbox_app_name, type_filter="DockerContainer"
    )
    # logger.debug(f"devbox_containers: {devbox_containers}")
    if devbox_containers is None:
        logger.error(f"DockerContainer:{devbox_app_name} not found")
        return False
    if len(devbox_containers) > 1:
        print_info(
            "Running commands in multiple containers is not yet supported. "
            + "Running in the first container. "
        )
    devbox_container: DockerContainer = devbox_containers[0]
    # logger.debug("devbox_container: ")
    # logger.debug("Name: {}".format(devbox_container.name))
    # logger.debug("Class: {}".format(devbox_container.__class__))
    # logger.debug("Resource: {}".format(devbox_container))
    docker_client: DockerApiClient = docker_manager.docker_worker.docker_client
    active_container: Optional[Container] = devbox_container.read(docker_client)
    # logger.debug("active_container: {}".format(active_container.attrs))
    # logger.debug("Class: {}".format(active_container.__class__))
    # logger.debug("Type: {}".format(type(active_container)))
    if active_container is None or not isinstance(active_container, Container):
        print_error("Container not available, please check your workspace is deployed.")
        return False

    # Step 4: Run the command
    print_subheading(f"Running command: `{command}`")
    run_status = execute_command(
        cmd=command,
        container=active_container,
        docker_env=docker_config.docker_env,
    )
    logger.debug(f"Run status: {run_status}")
    return run_status



from pathlib import Path
from typing import Optional, Dict, List

from phidata.product import DataProduct
from phidata.workflow import Workflow
from phidata.workspace import WorkspaceConfig
from phidata.types.context import RunContext
from pydantic import ValidationError

from phi.utils.cli_console import (
    print_error,
    print_info,
    print_subheading,
    print_heading,
    print_info,
    print_warning,
)
from phi.utils.log import logger
from phi.workflow.wf_enums import WorkflowEnv
from phi.workflow.wf_utils import (
    get_workflow_file_and_name,
    get_data_products_and_workflows_from_file,
)
from phi.workspace.phi_ws_data import PhiWsData


def run_workflow(
    wf_description: str,
    ws_data: PhiWsData,
    target_env: WorkflowEnv,
    target_dt: str,
    dry_run: bool,
    target_app: Optional[str] = None,
    run_env_vars: Optional[Dict[str, str]] = None,
    run_params: Optional[Dict[str, str]] = None,
) -> None:
    """Runs the Phidata Workflow in the target environment.
    The run_workflow() function is called from `phi wf run`
    """

    if ws_data is None or ws_data.ws_config is None:
        print_error("WorkspaceConfig invalid")
        return
    ws_config: WorkspaceConfig = ws_data.ws_config

    # Step 1: Get the ws_dir_path and products_dir
    if ws_data.ws_dir_path is None:
        print_error("Workspace directory invalid")
        return
    ws_dir_path: Path = ws_data.ws_dir_path

    if ws_config.products_dir is None:
        print_error("Products directory invalid")
        return
    products_dir: str = ws_config.products_dir

    # Step 2: Get the workflow_file and workflow_name
    # The workflow_file is the path of the file relative to the products_dir.
    workflow_file, workflow_name = get_workflow_file_and_name(
        wf_description=wf_description,
        ws_dir_path=ws_dir_path,
        products_dir=products_dir,
    )
    # Validate workflow_file
    if workflow_file is None:
        print_error("Workflow file not found")
        return
    print_heading(f"Running workflow: {wf_description}")
    print_info("Loading file: {}".format(workflow_file))
    if workflow_name:
        print_info("Workflow: {}".format(workflow_name))

    # Step 3: Get the absolute workflow_file_path
    workflow_file_path: Path
    try:
        workflow_file_path = (
            ws_dir_path.joinpath(products_dir)
            .joinpath(workflow_file)
            .resolve(strict=True)
        )
    except FileNotFoundError as e:
        print_error("Workflow file not found: {}".format(workflow_file))
        return

    # Step 4: Get DataProducts/Workflows in the workflow_file_path
    dp_dict: Dict[str, DataProduct]
    wf_dict: Dict[str, Workflow]
    try:
        dp_dict, wf_dict = get_data_products_and_workflows_from_file(workflow_file_path)
    except ValidationError as e:
        from phidata.utils.cli_console import print_validation_errors

        print_validation_errors(e.errors())
        print_info("Please fix and try again")
        return

    # Final run status
    run_status: bool = False
    # Build the RunContext
    run_context = RunContext(
        run_date=target_dt,
        dry_run=dry_run,
        run_env=target_env.value,
        run_env_vars=run_env_vars,
        run_params=run_params,
    )

    # Step 5: Run DataProducts/Workflows in target runtime
    # Run Workflow locally
    if target_env == WorkflowEnv.local:
        from phi.local.local_operator import run_workflows_local
        from phi.local.local_operator import run_data_products_local

        # ** CASE 1: If a workflow_name is provided
        #   Run only the Workflow
        if workflow_name is not None and len(wf_dict) > 0:
            run_status = run_workflows_local(
                workflow_file=workflow_file,
                workflows=wf_dict,
                run_context=run_context,
                ws_data=ws_data,
                workflow_name=workflow_name,
            )
        # ** CASE 2: If a workflow_name is not provided
        #   but workflow_file contains DataProducts.
        #   Run the DataProducts
        elif dp_dict is not None and len(dp_dict) > 0:
            run_status = run_data_products_local(
                workflow_file=workflow_file,
                data_products=dp_dict,
                run_context=run_context,
                ws_data=ws_data,
            )
        # ** CASE 3: If a workflow_name is not provided
        #   and workflow_file does NOT contain DataProducts
        #   but workflow_file contains Workflows
        #   Run the Workflows
        elif wf_dict is not None and len(wf_dict) > 0:
            run_status = run_workflows_local(
                workflow_file=workflow_file,
                workflows=wf_dict,
                run_context=run_context,
                ws_data=ws_data,
            )
    # Run Workflow in docker
    elif target_env == WorkflowEnv.dev:
        from phidata.infra.docker.config import DockerConfig
        from phi.docker.docker_operator import (
            run_workflows_docker,
            run_data_products_docker,
        )

        docker_configs: Optional[List[DockerConfig]] = ws_config.docker
        docker_config_to_use: Optional[DockerConfig] = None
        if docker_configs is not None and isinstance(docker_configs, list):
            if len(docker_configs) == 1:
                docker_config_to_use = docker_configs[0]
            else:
                for dc in docker_configs:
                    if dc.env == target_env:
                        docker_config_to_use = dc
                        break
        if docker_config_to_use is None:
            print_error(f"No DockerConfig found for env: {target_env.value}")
            return

        ######################################################################
        # NOTE: VERY IMPORTANT TO GET RIGHT
        # Update sub-configs data using WorkspaceConfig
        # 1. Pass down the paths from the WorkspaceConfig
        #       These paths are used everywhere from Infra to Apps
        # 2. Pass down docker_env which is used to set the env variables
        #       when running the docker command
        ######################################################################

        # The ws_dir_path is the ROOT directory for the workspace
        docker_config_to_use.workspace_root_path = ws_data.ws_dir_path
        docker_config_to_use.workspace_config_file_path = ws_data.ws_config_file_path
        docker_config_to_use.scripts_dir = ws_config.scripts_dir
        docker_config_to_use.storage_dir = ws_config.storage_dir
        docker_config_to_use.meta_dir = ws_config.meta_dir
        docker_config_to_use.products_dir = ws_config.products_dir
        docker_config_to_use.notebooks_dir = ws_config.notebooks_dir
        docker_config_to_use.workspace_config_dir = ws_config.workspace_config_dir
        docker_config_to_use.docker_env = ws_config.docker_env

        # ** CASE 1: If a workflow_name is provided
        #   Run only the Workflow
        if workflow_name is not None and len(wf_dict) > 0:
            # Because this is a single workflow, we need to identify the DAG
            # it belongs to. Workflow DAGs are usually provided by their DataProducts.
            # But Workflows can independently have their own DAGs as well.
            # To run a single workflow, lets find out if this workflow
            #   is a part of a DataProduct
            #   or an independent workflow
            wf_dag_id = None
            logger.debug(f"Finding DAG for {workflow_name}")
            if len(dp_dict) > 0:
                if workflow_name not in wf_dict:
                    print_error(
                        "Could not find '{}' in {}".format(
                            workflow_name, "[{}]".format(", ".join(wf_dict.keys()))
                        )
                    )
                else:
                    _workflow = wf_dict[workflow_name]
                    # logger.debug(f"Found workflow: {_workflow}")
                    for dp_name, dp_obj in dp_dict.items():
                        dp_workflows = dp_obj.workflows
                        # logger.debug(f"dp_workflows: {dp_workflows}")
                        if dp_workflows is not None and _workflow in dp_workflows:
                            # logger.debug(f"Found data product: {dp_name}")
                            wf_dag_id = dp_obj.dag_id
                            print_info(f"Workflow DAG: {wf_dag_id}")

            run_status = run_workflows_docker(
                workflow_file=workflow_file,
                workflows=wf_dict,
                run_context=run_context,
                docker_config=docker_config_to_use,
                target_app=target_app,
                workflow_name=workflow_name,
                use_dag_id=wf_dag_id,
            )
        # ** CASE 2: If a workflow_name is not provided
        #   but workflow_file contains DataProducts.
        #   Run the DataProducts
        elif dp_dict is not None and len(dp_dict) > 0:
            run_status = run_data_products_docker(
                workflow_file=workflow_file,
                data_products=dp_dict,
                run_context=run_context,
                docker_config=docker_config_to_use,
                target_app=target_app,
            )
        # ** CASE 3: If a workflow_name is not provided
        #   and workflow_file does NOT contain DataProducts
        #   but workflow_file contains Workflows
        #   Run the Workflows
        elif wf_dict is not None and len(wf_dict) > 0:
            run_status = run_workflows_docker(
                workflow_file=workflow_file,
                workflows=wf_dict,
                run_context=run_context,
                docker_config=docker_config_to_use,
                target_app=target_app,
            )
    # Run Workflow in staging k8s
    elif target_env == WorkflowEnv.stg:
        print_error(f"WorkflowEnv: {target_env} not yet supported")
    # Run Workflow in prod k8s
    elif target_env == WorkflowEnv.prd:
        print_error(f"WorkflowEnv: {target_env} not yet supported")
    else:
        print_error(f"WorkflowEnv: {target_env} not supported")

    # Report the run_status
    if run_status:
        print_heading("Workflow run success")
    else:
        print_error("Workflow run failure")


def list_workflows(
    wf_description: str,
    ws_data: PhiWsData,
) -> None:

    if ws_data is None or ws_data.ws_config is None:
        print_error("WorkspaceConfig invalid")
        return
    ws_config: WorkspaceConfig = ws_data.ws_config

    # Step 1: Get the ws_dir_path and products_dir
    if ws_data.ws_dir_path is None:
        print_error("Workspace directory invalid")
        return
    ws_dir_path: Path = ws_data.ws_dir_path

    if ws_config.products_dir is None:
        print_error("Products directory invalid")
        return
    products_dir: str = ws_config.products_dir

    # Step 2: Get the workflow_file and workflow_name
    # The workflow_file is the path of the file relative to the products_dir.
    workflow_file, workflow_name = get_workflow_file_and_name(
        wf_description=wf_description,
        ws_dir_path=ws_dir_path,
        products_dir=products_dir,
    )
    # Validate workflow_file
    if workflow_file is None:
        print_error("Workflow file not found")
        return
    print_heading(f"Workflows in: {wf_description}")
    print_info("Loading file: {}".format(workflow_file))
    if workflow_name:
        print_info("Workflow: {}".format(workflow_name))

    # Step 3: Get the absolute workflow_file_path
    workflow_file_path: Path
    try:
        workflow_file_path = (
            ws_dir_path.joinpath(products_dir)
            .joinpath(workflow_file)
            .resolve(strict=True)
        )
    except FileNotFoundError as e:
        print_error("Workflow file not found: {}".format(workflow_file))
        return

    # Step 4: Get DataProducts/Workflows in the workflow_file_path
    dp_dict: Dict[str, DataProduct]
    wf_dict: Dict[str, Workflow]
    try:
        dp_dict, wf_dict = get_data_products_and_workflows_from_file(workflow_file_path)
    except ValidationError as e:
        from phidata.utils.cli_console import print_validation_errors

        print_validation_errors(e.errors())
        print_info("Please fix and try again")
        return

    # Step 5: Print :)
    if dp_dict is not None and len(dp_dict) > 0:
        print_heading(f"\nFound {len(dp_dict)} DataProduct(s)")
        for key, dp in dp_dict.items():
            print_info("{}: {}".format(key, dp.name))
    if wf_dict is not None and len(wf_dict) > 0:
        print_heading(f"\nFound {len(wf_dict)} Workflow(s)")
        for key, wf in wf_dict.items():
            print_info("{}: {}".format(key, wf.name))
        if workflow_name is not None:
            if workflow_name not in wf_dict:
                print_info("")
                print_warning(
                    "Could not find '{}' in {}".format(
                        workflow_name, "[{}]".format(", ".join(wf_dict.keys()))
                    )
                )



from typing import List, Optional

from typer import launch as typer_launch

from phi.backend_api.api_exceptions import CliAuthException
from phi.backend_api.auth_api import (
    authenticate_and_get_user_schema,
    is_user_authenticated,
    sign_in_user,
)
from phi.backend_api.workspace_api import (
    get_available_workspaces,
    get_primary_workspace,
)
from phi.conf.constants import PHI_CONF_DIR, PHI_SIGNIN_URL_WITHOUT_PARAMS
from phi.conf.phi_conf import PhiConf
from phi.schemas.auth_schemas import EmailPasswordSignInSchema
from phi.schemas.user_schemas import UserSchema
from phi.utils.cli_auth_server import (
    get_port_for_auth_server,
    get_tmp_access_token_from_web_flow,
)
from phi.utils.cli_console import (
    print_info,
    print_error,
    print_heading,
    print_info,
    print_info,
    print_subheading,
)
from phi.utils.common import is_empty
from phi.utils.filesystem import delete_from_fs
from phi.utils.log import logger
from phi.workspace.ws_operator import init_workspace_data
from phi.workspace.ws_schemas import WorkspaceSchema


def delete_phidata_conf() -> None:
    logger.debug("Removing existing Phidata configuration")
    delete_from_fs(PHI_CONF_DIR)


def initialize_phidata(reset: bool = False) -> bool:
    """This function is called by `phi init` and initializes phidata on the users machine.
    Steps:
    1. Check if PHI_CONF_DIR exists, if not, create it. If reset == True, recreate PHI_CONF_DIR.
    2. Check if PhiConf exists, if it does, try and authenticate the user
        If the user is authenticated, phi is configured and authenticated. Return True.
    3. If PhiConf does not exist, create a new PhiConf. Return True.
    """

    print_heading("Welcome to phidata!\n")
    if reset:
        delete_phidata_conf()

    logger.debug("Initializing phidata")

    # Check if ~/.phi exists, if its not a dir - delete it and create the dir
    if PHI_CONF_DIR.exists():
        # logger.debug(f"{PHI_CONF_DIR} exists")
        if not PHI_CONF_DIR.is_dir():
            try:
                delete_from_fs(PHI_CONF_DIR)
            except Exception as e:
                logger.exception(e)
                print_info(
                    f"Something went wrong, please delete {PHI_CONF_DIR} and run `phi init` again"
                )
                return False
            PHI_CONF_DIR.mkdir(parents=True)
    else:
        # logger.debug(f"creating {PHI_CONF_DIR}")
        PHI_CONF_DIR.mkdir(parents=True)
        logger.debug(f"created {PHI_CONF_DIR}")

    # Confirm PHI_CONF_DIR exists otherwise we should return
    if PHI_CONF_DIR.exists():
        logger.debug(f"Your phidata config is stored at: {PHI_CONF_DIR}")
    else:
        print_info(f"Something went wrong, please run `phi init` again")
        return False

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if phi_conf is None:
        # Create a new PhiConf
        phi_conf = PhiConf()

    # Authenticate user
    is_authenticated = is_user_authenticated()
    if is_authenticated:
        logger.debug(f"Phidata is configured and authenticated")
        return True

    if phi_conf is not None:
        print_info("Phidata is initialized. Next steps:\n")
        print_info(" - Run `phi ws init` to create a new workspace")
        print_info(
            " - Run `phi ws setup` from an existing directory to setup the workspace"
        )
        print_info(" - Run `phi ws up -dr` to dry-run workspace deploy")
        print_info(" - Run `phi ws up` to deploy the workspace")
    return True


def authenticate_phidata() -> bool:
    """Authenticate the user with the phidata backend.
    Steps:
    1. Authenticate the user by opening the phidata sign-in url
        and the web-app will post back a temporary CLI auth token
        to our mini http server running on auth_server_port.
    2. Using the tmp_access_token, authenticate the CLI with backend and
        save the phidata_session_cookie to PHI_CREDS_PATH.
        This step is handled by authenticate_and_get_user_schema()
    5. Once the user is authenticated, create a PhiConf for ths user.
    6. Get the primary and available workspaces for this user.
    7. If the user has workspaces available, initialize additional workspace requirements
    """

    print_heading("Authenticating with phidata")

    auth_server_port = get_port_for_auth_server()
    redirect_uri = "http%3A%2F%2Flocalhost%3A{}%2F".format(auth_server_port)
    auth_url = "{}?source=cli&action=INIT_CLI_AUTH&redirect_uri={}".format(
        PHI_SIGNIN_URL_WITHOUT_PARAMS, redirect_uri
    )
    print_info("\nYour browser will be opened to visit:\n{}".format(auth_url))
    typer_launch(auth_url)
    print_info("\nWaiting for a response from browser...\n")

    tmp_access_token = get_tmp_access_token_from_web_flow(auth_server_port)
    if tmp_access_token is None:
        print_error(f"Could not authenticate, please run `phi auth` again")
        return False

    try:
        user: Optional[UserSchema] = authenticate_and_get_user_schema(tmp_access_token)
    except CliAuthException as e:
        logger.exception(e)
        print_error(f"Could not authenticate, please run `phi auth` again")
        return False

    if user is None:
        print_error(f"Could not retrieve user data, please run `phi auth` again")
        return False

    print_info("Welcome {}, you are now authenticated\n".format(user.username))

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if phi_conf is None:
        phi_conf = PhiConf()

    phi_conf.user = user
    # Get Primary/Available Workspaces for the user
    active_ws_name: Optional[str] = phi_conf.active_ws_name
    available_workspaces: Optional[List[WorkspaceSchema]] = phi_conf.available_ws

    # If the active_ws_name is None, call backend to get the primary_ws
    if active_ws_name is None:
        logger.debug("Getting primary workspace from api")
        primary_ws = get_primary_workspace()
        if primary_ws is not None:
            logger.debug(
                "Received primary workspace:\n{}".format(primary_ws.json(indent=2))
            )
            phi_conf.active_ws_name = primary_ws.name
        else:
            logger.debug("No primary workspace available")

    # If available_workspaces is empty and a primary_ws is available
    # call backend to get the available workspaces
    if is_empty(available_workspaces) and active_ws_name is not None:
        logger.debug("Getting available workspaces from api")
        available_workspaces = get_available_workspaces()
        if available_workspaces:
            logger.debug(
                "Received {} available workspace(s)".format(len(available_workspaces))
            )
            phi_conf.available_ws = available_workspaces
            for _ws in available_workspaces:
                init_workspace_data(_ws, phi_conf)

    return True


def sign_in_using_cli() -> bool:
    from getpass import getpass

    print_heading("Sign in")
    email_raw = input("email: ")
    pass_raw = getpass()

    if email_raw is None or pass_raw is None:
        print_error("incorrect email or password")

    try:
        user: Optional[UserSchema] = sign_in_user(
            EmailPasswordSignInSchema(email=email_raw, password=pass_raw)
        )
    except CliAuthException as e:
        logger.exception(e)
        return False

    if user is None:
        print_error("Could not get user data, sign in again")
        return False

    print_info("Welcome {}, you are now authenticated\n".format(user.username))

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if phi_conf is None:
        phi_conf = PhiConf()

    phi_conf.user = user
    # Get Primary/Available Workspaces for the user
    active_ws_name: Optional[str] = phi_conf.active_ws_name
    available_workspaces: Optional[List[WorkspaceSchema]] = phi_conf.available_ws

    # If the active_ws_name is None, call backend to get the primary_ws
    if active_ws_name is None:
        logger.debug("Getting primary workspace from api")
        primary_ws = get_primary_workspace()
        if primary_ws is not None:
            logger.debug(
                "Received primary workspace:\n{}".format(primary_ws.json(indent=2))
            )
            phi_conf.active_ws_name = primary_ws.name
        else:
            logger.debug("No primary workspace available")

    # If available_workspaces is empty and a primary_ws is available
    # call backend to get the available workspaces
    if is_empty(available_workspaces) and active_ws_name is not None:
        logger.debug("Getting available workspaces from api")
        available_workspaces = get_available_workspaces()
        if available_workspaces:
            logger.debug(
                "Received {} available workspace(s)".format(len(available_workspaces))
            )
            phi_conf.available_ws = available_workspaces
            for _ws in available_workspaces:
                init_workspace_data(_ws, phi_conf)

    return True



from pathlib import Path
from typing import Dict, Optional, Set, Tuple

from phi.backend_api.api_exceptions import CliAuthException
from phi.backend_api.auth_api import authenticate_and_get_user_schema
from phi.conf.constants import PHI_SIGNIN_URL_WITHOUT_PARAMS
from phi.conf.phi_conf import PhiWsData
from phi.schemas.user_schemas import UserSchema
from phi.utils.cli_auth_server import (
    get_port_for_auth_server,
    get_tmp_access_token_from_web_flow,
)
from phi.utils.cli_console import (
    print_info,
    print_info,
    print_error,
    print_subheading,
)
from phi.utils.common import is_empty
from phi.utils.log import logger
from phi.workspace.ws_enums import WorkspaceSetupActions
from phi.workspace.ws_exceptions import WorkspaceConfigException


def get_ws_config_file_path(ws_dir: Path) -> Path:
    logger.debug(f"Looking for a workspace config for {ws_dir}")
    # Phidata looks for a workspace config dir in 2 places:
    # 1. In a folder with the same name as ws_dir
    # 2. In a folder named data
    #
    # When users have multiple workspaces in the same env, its better to have
    # different names for each package to separate the imports. So the package/folder
    # name is the same name as the workspace
    #
    # But in 99% of the cases, users will have 1 workspace per virtual env, so
    # using the workspace name: data works as a great default option

    ws_config_dir: Optional[Path] = None

    # Case 1: Look for a folder with the same name as ws_dir
    ws_root_name = ws_dir.stem
    ws_src_dir = ws_dir.joinpath(ws_root_name)
    if ws_src_dir.exists() and ws_src_dir.is_dir():
        ws_config_dir = ws_src_dir.joinpath("workspace")
    # Case 2: Look for a folder with the name: data
    if ws_config_dir is None:
        ws_data_dir = ws_dir.joinpath("data")
        if ws_data_dir.exists() and ws_data_dir.is_dir():
            ws_config_dir = ws_data_dir.joinpath("workspace")

    if ws_config_dir is not None and ws_config_dir.exists() and ws_config_dir.is_dir():
        return ws_config_dir.joinpath("config.py")
    raise WorkspaceConfigException(f"Could not find a workspace config in {ws_dir}")


def get_ws_k8s_resources_dir_path(ws_dir: Path) -> Path:
    return ws_dir.joinpath("workspace").joinpath("prd")


def is_valid_ws_config_file_path(ws_config_file_path: Optional[Path]) -> bool:
    # TODO: add more checks for validating the ws_config_file_path
    if (
        ws_config_file_path is not None
        and ws_config_file_path.exists()
        and ws_config_file_path.is_file()
    ):
        return True
    return False


def select_version_contol_using_web_flow() -> Optional[UserSchema]:
    import typer

    port = get_port_for_auth_server()
    redirect_uri = "http%3A%2F%2Flocalhost%3A{}%2F".format(port)
    auth_url = "{}?source=cli&action=SELECT_VERSION_CONTROL&redirect_uri={}".format(
        PHI_SIGNIN_URL_WITHOUT_PARAMS, redirect_uri
    )
    print_info("\nYour browser will be opened to visit:\n\n{}\n".format(auth_url))
    typer.launch(auth_url)
    print_info("Waiting for a response from browser")

    tmp_access_token = get_tmp_access_token_from_web_flow(port)
    if tmp_access_token is None:
        print_error("tmp_access_token")
        return None
    try:
        user: Optional[UserSchema] = authenticate_and_get_user_schema(tmp_access_token)
    except CliAuthException as e:
        print_error("CliAuthException")
        return None

    return user


def get_ws_setup_status(ws_data: PhiWsData) -> Tuple[bool, str]:
    """This function validates that a workspace is properly set up and returns an
    error message if ws not setup. The error message is displayed to the user,
    so we need to make sure it doesn't contain any sensitive information.
    """

    # Version Control Provider check
    # user: UserSchema = config.user
    # if user.version_control_provider is None:
    #     return (
    #         False,
    #         "Version control provider not available. Please run `phi init -r` to initialize phi",
    #     )

    if ws_data is None:
        return (
            False,
            "WorkspaceSchema not yet registered with Phidata. Please run `phi ws init` to create a new workspace or `phi ws setup` for an existing workspace",
        )

    # Validate WorkspaceSchema Directory
    ws_dir_path: Optional[Path] = ws_data.ws_dir_path
    if ws_dir_path is None or not ws_dir_path.exists() or not ws_dir_path.is_dir():
        return (
            False,
            "The WorkspaceSchema directory is not available or invalid.\n\tTo create a new workspace, run `phi ws init`.\n\tFor an existing workspace, run `phi ws setup` from the workspace directory",
        )

    # Validate WorkspaceSchema Config Exists (we will validate the contents later)
    ws_config_file_path: Optional[Path] = ws_data.ws_config_file_path
    if (
        ws_config_file_path is None
        or not ws_config_file_path.exists()
        or not ws_config_file_path.is_file()
    ):
        return (
            False,
            "WorkspaceSchema config is not available or invalid. Please run `phi ws init` to create a new workspace or `phi ws setup` for an existing workspace",
        )

    # Validate that the workspace is registered with phidata
    if ws_data is not None and ws_data.ws_schema is None:
        return (
            False,
            "WorkspaceSchema not registered with phidata. Please run `phi ws setup` from the workspace dir",
        )

    # Validate that the workspace has a git repo
    if (
        ws_data is not None
        and ws_data.ws_schema is not None
        and is_empty(ws_data.ws_schema.git_url)
    ):
        return (
            False,
            "WorkspaceSchema does not have a remote origin setup. Please run `phi ws setup` from the workspace dir",
        )

    return (True, "WorkspaceSchema setup successful")


ws_setup_action_todo: Dict[WorkspaceSetupActions, str] = {
    WorkspaceSetupActions.WS_CONFIG_IS_AVL: "[ERROR] WorkspaceConfig is missing",
    WorkspaceSetupActions.WS_IS_AUTHENTICATED: "Workspace is not authenticated. run `phi auth` for more info",
    WorkspaceSetupActions.GCP_SVC_ACCOUNT_IS_AVL: "GCP Service Account is unavailable. run `phi gcp auth` for more info",
    WorkspaceSetupActions.GIT_REMOTE_ORIGIN_IS_AVL: "No git repo setup for workspace. run `phi ws git` for more info",
}


def print_howtofix_pending_actions(
    pending_actions: Optional[Set[WorkspaceSetupActions]],
) -> None:
    """This function prints how to fix pending setup actions for a workspace"""

    if pending_actions is None:
        return

    for action in pending_actions:
        print_info("\t" + ws_setup_action_todo.get(action, action.value))


def secho_ws_status(ws_data: PhiWsData) -> None:

    ws_dir_path: Optional[Path] = ws_data.ws_dir_path
    ws_config_file_path: Optional[Path] = ws_data.ws_config_file_path

    print_info("")
    print_subheading("WorkspaceSchema: {}".format(ws_data.ws_name))
    print_info("WorkspaceSchema Directory: {}".format(str(ws_dir_path)))
    print_info("Workspace Config file: {}".format(str(ws_config_file_path)))

    ws_is_setup, ws_setup_msg = get_ws_setup_status(ws_data)
    if ws_is_setup:
        print_info("WorkspaceSchema Status: Active")
    else:
        print_info("WorkspaceSchema Status: {}".format(ws_setup_msg))

    # ws_pak8_conf: Optional[Pak8Conf] = ws_data.ws_pak8_conf
    # if ws_pak8_conf and ws_pak8_conf.cloud_provider == pak8_Pak8CloudProvider.GCP:
    #     secho_gcp_status(ws_data)



from collections import OrderedDict
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple

from phi.conf.constants import PHI_CONF_PATH
from phi.schemas.user_schemas import UserSchema
from phi.utils.cli_console import (
    print_available_workspaces,
    print_info,
    print_dict,
    print_error,
    print_heading,
    print_horizontal_line,
    print_info,
    print_subheading,
)
from phi.utils.common import pickle_object_to_file, unpickle_object_from_file
from phi.utils.dttm import current_datetime_utc
from phi.utils.filesystem import delete_from_fs
from phi.utils.log import logger
from phi.workspace.phi_ws_data import PhiWsData
from phi.workspace.ws_enums import WorkspaceSetupActions
from phi.workspace.ws_schemas import WorkspaceSchema


class PhiConf:
    """The PhiConf class manages data for the phi cli"""

    def __init__(self):
        # Current user, populated after authenticating with the backend
        # To add a user, use the user setter
        self._user: Optional[UserSchema] = None

        # Primary ws for the user - used as the default for `phi ws` commands
        # To add a primary workspace, use the active_ws_name setter
        self._active_ws_name: Optional[str] = None

        # ** The self._ws_data_map holds the data for all workspaces available to this user.
        self._ws_data_map: Dict[str, PhiWsData] = OrderedDict()
        # When phi is initialized using `phi init`, this is an empty dict
        # When the user is authenticated, we get a list of all the workspaces this user has access to.
        # These workspaces are then added using the `primary_ws` and `available_ws` setters.
        # Note: These workspaces arent yet cloned on the users machines and only a record of their availability is kept
        # When the workspace is setup on the users machine, we update the remaining workspace data.

        # Quick access from ws_path -> ws data.
        self._path_to_ws_data_map: Dict[Path, PhiWsData] = OrderedDict()

        # Save the config to disk once initialized
        self.save_config()

    ######################################################
    ## User functions
    ######################################################

    @property
    def user(self) -> Optional[UserSchema]:
        return self._user

    @user.setter
    def user(self, user: Optional[UserSchema]) -> None:
        """Sets the user"""
        if user is not None:
            self._user = user
            self.save_config()

    ######################################################
    ## Workspace functions
    ######################################################

    @property
    def active_ws_name(self) -> Optional[str]:
        return self._active_ws_name

    @active_ws_name.setter
    def active_ws_name(self, ws_name: Optional[str]) -> None:
        """Set the active_ws_name during phi auth"""
        if ws_name is not None:
            self._active_ws_name = ws_name
            self._add_or_update_ws_data(ws_name=ws_name)
            self.save_config()

    @property
    def available_ws(self) -> List[WorkspaceSchema]:
        return [
            w.ws_schema for w in self._ws_data_map.values() if w.ws_schema is not None
        ]

    @available_ws.setter
    def available_ws(self, avl_ws: Optional[List[WorkspaceSchema]]) -> None:
        """Set the available ws during phi init"""
        if avl_ws:
            for ws_schema in avl_ws:
                self._add_or_update_ws_data(ws_name=ws_schema.name, ws_schema=ws_schema)
            self.save_config()

    def _add_or_update_ws_data(
        self,
        ws_name: str,
        ws_schema: Optional[WorkspaceSchema] = None,
        ws_dir_path: Optional[Path] = None,
        ws_config_file_path: Optional[Path] = None,
        refresh_data: bool = False,
    ) -> None:
        """Create or update PhiWsData in PhiConf.
        ws_name is the only required argument.
        This function does not call self.save_config(). call it from the parent.

        This function is called from 5 places:
        1. During `phi auth` when primary/available workspaces are requested from the backend.
        2. During `phi ws init` when a new workspace is created on the users machine.
        3. When the user already has a local workspace dir on their machine
            and need to manually add the ws to the PhiConf
            (possible after running `phi init -r` which reset the config)
        4. Whenever ws config, creds or data needs to be refreshed. This function is called with just the `ws_name` arg
        5. When primary or available workspaces are updated, this function is called from the
            @primary_ws.setter and @available_ws.setter with the ws_name and ws_schema

        Note: Cloud data (ws_gcp_data etc.) should be updated using their respective functions.
        """

        ######################################################
        # Validate ws_name and ws_schema
        ######################################################

        # ws_name should not be None
        if ws_name is None or not isinstance(ws_name, str):
            return

        ######################################################
        # Create ws_data
        ######################################################

        if ws_name not in self._ws_data_map:
            logger.debug(f"Adding new ws {ws_name} to PhiConf")
            self._ws_data_map[ws_name] = PhiWsData(
                ws_name=ws_name,
                ws_schema=ws_schema,
                ws_dir_path=ws_dir_path,
                ws_config_file_path=ws_config_file_path,
            )
            if ws_dir_path:
                self._path_to_ws_data_map[ws_dir_path] = self._ws_data_map[ws_name]
            logger.debug(
                "Created new Workspace: {}".format(
                    self._ws_data_map[ws_name].json(indent=2)
                )
            )

        ######################################################
        # Update or Refresh ws_data
        ######################################################

        # By this point there should be a PhiWsData object for this ws_name
        # If by chance ws_name not in self._ws_data_map, then we just create a new PhiWsData object
        existing_ws_data: PhiWsData = self._ws_data_map.get(
            ws_name, PhiWsData(ws_name=ws_name)
        )
        # Make a deep copy of the existing workspace data
        # This allows us to update the fields we want, and keep any existing fields as is
        _ws_data: PhiWsData = existing_ws_data.copy(deep=True)

        # Add values where available
        if ws_schema is not None and ws_schema != existing_ws_data.ws_schema:
            logger.info("Updating ws_schema")
            _ws_data.ws_schema = ws_schema
        if ws_dir_path is not None and ws_dir_path != existing_ws_data.ws_dir_path:
            logger.info("Updating ws_dir_path")
            _ws_data.ws_dir_path = ws_dir_path
        if (
            ws_config_file_path is not None
            and ws_config_file_path != existing_ws_data.ws_config_file_path
        ):
            logger.info("Updating ws_config_file_path")
            _ws_data.ws_config_file_path = ws_config_file_path

        # Refresh the WorkspaceConfig if needed.
        # There are 3 cases where we read/parse the WorkspaceConfig
        # 1. refresh_data == True
        # 2. ws_config_file_path has changed
        # 3. current ws_config is None
        # But before we refresh the WorkspaceConfig, we need to check if a ws_config_file_path is available
        if _ws_data.ws_config_file_path is not None:
            if (
                refresh_data
                or _ws_data.ws_config_file_path != existing_ws_data.ws_config_file_path
                or _ws_data.ws_config is None
            ):
                _ws_data.load_workspace_from_config_file()
        _ws_data.last_update_ts = current_datetime_utc()

        # Point the ws_data in _ws_data_map and _path_to_ws_data_map to _ws_data
        logger.debug(f"Updating ws_data_map object for: {ws_name}")
        # Pop the existing object from the self._ws_data_map
        if ws_name in self._ws_data_map:
            popped_ws_data_from_map: Optional[PhiWsData] = self._ws_data_map.pop(
                ws_name
            )
            # logger.debug(f"Removed existing ws_data for {ws_name} from mapping")
        self._ws_data_map[ws_name] = _ws_data

        if _ws_data.ws_dir_path is not None:
            # If the ws_dir_path has been updated, pop the existing_ws_data.ws_dir_path
            if (
                existing_ws_data.ws_dir_path is not None
                and existing_ws_data.ws_dir_path in self._path_to_ws_data_map
            ):
                popped_ws_data_from_path_map: Optional[
                    PhiWsData
                ] = self._path_to_ws_data_map.pop(existing_ws_data.ws_dir_path)
            self._path_to_ws_data_map[_ws_data.ws_dir_path] = self._ws_data_map[ws_name]

        ######################################################
        # END
        ######################################################

    def add_new_created_ws_to_config(
        self,
        ws_name: str,
        ws_dir_path: Path,
        ws_config_file_path: Path,
        refresh_data: bool = False,
    ) -> None:
        """Adds a newly created workspace to the PhiConf"""
        self._add_or_update_ws_data(
            ws_name=ws_name,
            ws_dir_path=ws_dir_path,
            ws_config_file_path=ws_config_file_path,
            refresh_data=refresh_data,
        )
        self.save_config()

    def update_ws_schema_for_workspace(
        self,
        ws_schema: WorkspaceSchema,
        set_as_primary: bool = False,
    ) -> None:
        """Updates the schema for a workspace and sets as primary if needed"""
        self._add_or_update_ws_data(
            ws_name=ws_schema.name,
            ws_schema=ws_schema,
        )
        if set_as_primary:
            self.active_ws_name = ws_schema.name
        self.save_config()

    def update_ws_data(
        self,
        ws_name: str,
        ws_schema: Optional[WorkspaceSchema] = None,
        ws_dir_path: Optional[Path] = None,
        ws_config_file_path: Optional[Path] = None,
    ) -> None:
        """Update workspace data and return True if successful"""

        print_info(f"Updating ws_data for {ws_name}")
        self._add_or_update_ws_data(
            ws_name=ws_name,
            ws_schema=ws_schema,
            ws_dir_path=ws_dir_path,
            ws_config_file_path=ws_config_file_path,
        )
        self.save_config()

    def update_ws_data_for_cloned_ws(
        self,
        ws_name: str,
        ws_dir_path: Path,
        ws_config_file_path: Path,
    ) -> bool:
        """For a workspace to be cloned, it would already have ws_data and ws_schema available.
        After the directory from the git_url is cloned, we can fill in the gaps to update
        the ws_data for the workspace. Call this function through ws_operator.clone_workspace()
        """
        ws_data: Optional[PhiWsData] = self._ws_data_map.get(ws_name, None)
        if ws_data is None:
            return False

        self._add_or_update_ws_data(
            ws_name=ws_name,
            ws_dir_path=ws_dir_path,
            ws_config_file_path=ws_config_file_path,
        )
        self.save_config()
        return True

    def add_ws_data_using_config_file(
        self, ws_dir_path: Path, ws_config_file_path: Optional[Path]
    ) -> Optional[PhiWsData]:
        """Maps a ws at `ws_dir_path` to an available workspace. using the config file at `ws_config_file_path`.
        Also updates the ws_config_file_path in the ws_data if provided.

        This is used in 2 cases:
        1. If the user ran `phi init -r`, the PhiConf gets erased and so does self._path_to_ws_data_map
            So there is no link from the ws_dir_path to the workspace.
        2. The user manually cloned the workspace directory. In that case we don't have a record of this
            ws_dir_path matching any available workspace.

        This function basically reads the workspace config and tries to map it to an available workspace.
        Returns the PhiWsData if successful
        """

        logger.debug(f"Looking for the workspace at {ws_dir_path}")
        _best_guess_ws_name = ws_dir_path.stem
        logger.debug(f"_best_guess_ws_name: {_best_guess_ws_name}")
        _ws_schema: Optional[WorkspaceSchema] = None
        for avl_ws in self.available_ws:
            if avl_ws.name == _best_guess_ws_name:
                _ws_schema = avl_ws
            break
        if _ws_schema is None:
            print_error(
                f"Workspace name: {_best_guess_ws_name} does not match any available workspaces"
            )
            print_available_workspaces(self.available_ws)
            return None

        logger.debug(f"Found matching workspace {_ws_schema.name}")
        logger.debug(
            f"Mapping {_ws_schema.name} to dir: {ws_dir_path} and config: {ws_config_file_path}"
        )
        self._add_or_update_ws_data(
            ws_name=_ws_schema.name,
            ws_schema=_ws_schema,
            ws_dir_path=ws_dir_path,
            ws_config_file_path=ws_config_file_path,
            refresh_data=True,
        )
        self.save_config()
        return self._ws_data_map.get(_ws_schema.name, None)

    def validate_workspace(
        self, ws_name: str
    ) -> Tuple[bool, Optional[Set[WorkspaceSetupActions]]]:
        """Validates a workspace and returns a tuple of [validation_status, pending_setup_actions]
        Returns:
            (False, None): If PhiWsData is not valid
            (True, None): If Validation is successful and everything is properly setup
            (True, set(WorkspaceSetupActions)): If Validation is successful but user has pending setup actions
            (False, set(WorkspaceSetupActions)): If Validation is unsuccessful and user has pending setup actions
        """

        # Validate the ws has a valid entry
        if ws_name is None or ws_name not in self._ws_data_map:
            return False, None
        logger.debug(f"Validating workspace: {ws_name}")

        ws_data: PhiWsData = self._ws_data_map[ws_name]

        # If ws_config is missing add a pending action
        if ws_data.ws_config is None:
            ws_data.pending_actions.add(WorkspaceSetupActions.WS_CONFIG_IS_AVL)
        else:
            ws_data.pending_actions.discard(WorkspaceSetupActions.WS_CONFIG_IS_AVL)

        # If ws_schema is missing add a pending action
        if ws_data.ws_schema is None:
            ws_data.pending_actions.add(WorkspaceSetupActions.WS_IS_AUTHENTICATED)
        else:
            ws_data.pending_actions.discard(WorkspaceSetupActions.WS_IS_AUTHENTICATED)

        # If git_url is missing add a pending action
        if ws_data.ws_schema is None or ws_data.ws_schema.git_url is None:
            ws_data.pending_actions.add(WorkspaceSetupActions.GIT_REMOTE_ORIGIN_IS_AVL)
        else:
            ws_data.pending_actions.discard(
                WorkspaceSetupActions.GIT_REMOTE_ORIGIN_IS_AVL
            )

        # logger.debug(f"Required Actions for this workspace: {ws_data.required_actions}")
        # logger.debug(f"Pending Actions for this workspace: {ws_data.pending_actions}")

        # Check how many required_actions are pending
        if len(ws_data.required_actions.intersection(ws_data.pending_actions)) > 0:
            return False, ws_data.pending_actions

        return True, ws_data.pending_actions

    def delete_ws(self, ws_dir_path: Path, delete_ws_dir: bool = False) -> bool:
        """Handles Deleting a workspace from the PhiConf using ws_dir_path"""

        ws_data: Optional[PhiWsData] = None
        if ws_dir_path in self._path_to_ws_data_map:
            ws_data = self._path_to_ws_data_map.pop(ws_dir_path, None)
        # Return False if a ws using this path does not exist
        if ws_data is None:
            print_error(f"Could not locate workspace on path {str(ws_dir_path)}")
            return False

        print_info("")
        print_info(f"Deleting workspace {ws_data.ws_name}")
        print_info("")
        # Check if we're deleting the primary workspace, if yes, unset the primary ws
        if (
            self._active_ws_name is not None
            and ws_data.ws_name is not None
            and self._active_ws_name == ws_data.ws_name
        ):
            print_info(f"Removing {ws_data.ws_name} as the primary workspace")
            self._active_ws_name = None

        # Then delete the workspace data
        logger.debug(f"Removing WorkspaceSchema Data")
        deleted_ws_data: Optional[PhiWsData] = self._ws_data_map.pop(
            ws_data.ws_name, None
        )

        ws_dir_path_deleted: bool = False
        if delete_ws_dir and ws_dir_path:
            print_info(f"Deleting {str(ws_dir_path)}")
            ws_dir_path_deleted = delete_from_fs(ws_dir_path)

        # Save the config after deleting workspace
        self.save_config()

        if deleted_ws_data is not None:
            if delete_ws_dir:
                return ws_dir_path_deleted
            return True
        return False

    ######################################################
    ## Get Workspace Data
    ######################################################

    ## Workspace Name
    def get_ws_name_by_path(self, ws_dir_path: Path) -> Optional[str]:
        if ws_dir_path in self._path_to_ws_data_map:
            return self._path_to_ws_data_map[ws_dir_path].ws_name
        return None

    ## Workspace Schema
    def get_ws_schema_by_name(self, ws_name: str) -> Optional[WorkspaceSchema]:
        if ws_name in self._ws_data_map:
            return self._ws_data_map[ws_name].ws_schema
        return None

    def get_ws_schema_by_path(self, ws_dir_path: Path) -> Optional[WorkspaceSchema]:
        if ws_dir_path in self._path_to_ws_data_map:
            return self._path_to_ws_data_map[ws_dir_path].ws_schema
        return None

    ## Workspace Directory
    def get_ws_dir_path_by_name(self, ws_name: str) -> Optional[Path]:
        if ws_name in self._ws_data_map:
            return self._ws_data_map[ws_name].ws_dir_path
        return None

    ## Workspace Config File Path
    def get_ws_config_file_path_by_name(self, ws_name: str) -> Optional[Path]:
        if ws_name in self._ws_data_map:
            return self._ws_data_map[ws_name].ws_config_file_path
        return None

    ## Workspace Pending Actions
    def get_ws_pending_actions_by_name(
        self, ws_name: str
    ) -> Optional[Set[WorkspaceSetupActions]]:
        if ws_name in self._ws_data_map:
            return self._ws_data_map[ws_name].pending_actions
        return None

    ## Workspace Data
    def get_ws_data_by_name(
        self, ws_name: str, refresh: bool = False
    ) -> Optional[PhiWsData]:
        if ws_name in self._ws_data_map:
            # if refresh or self._ws_data_map[ws_name].ws_config is None:
            if refresh:
                self.refresh_ws_config(ws_name)
            return self._ws_data_map[ws_name]
        return None

    def get_ws_data_by_path(self, ws_dir_path: Path) -> Optional[PhiWsData]:
        if ws_dir_path in self._path_to_ws_data_map:
            return self._path_to_ws_data_map[ws_dir_path]
        return None

    def refresh_ws_config(self, ws_name: str) -> bool:
        """Refresh the workspace config for a given workspace name"""

        logger.debug(f"++**++ Refreshing WorkspaceConfig for {ws_name}")
        self._add_or_update_ws_data(ws_name=ws_name, refresh_data=True)
        # we check that refresh was successful to avoid saving bad data
        # TODO: add checks to ensure refresh success.
        refresh_success = True
        if refresh_success:
            logger.debug(f"++**++ Refresh success")
            self.save_config()
        return refresh_success

    # ## Workspace Pak8Conf
    # def get_ws_pak8_conf_by_name(
    #     self, ws_name: str, refresh: bool = False, add_kubeconfig: bool = False
    # ) -> Optional[Pak8Conf]:
    #     if ws_name in self._ws_data_map:
    #         logger.debug(f"Getting Pak8Conf for {ws_name}")
    #         _ws_pak8_conf = self._ws_data_map[ws_name].ws_pak8_conf
    #         if refresh or _ws_pak8_conf is None:
    #             logger.debug(f"Refreshing Pak8Conf for {ws_name}")
    #             self.refresh_ws_config(ws_name)
    #         if add_kubeconfig:
    #             # Check if a kubeconfig is available, if yes, update the Pak8Conf
    #             ws_k8s_data: Optional[PhiK8sData] = self._ws_data_map[
    #                 ws_name
    #             ].ws_k8s_data
    #             if ws_k8s_data:
    #                 logger.debug(f"Adding Kubeconfig to Pak8Conf for {ws_name}")
    #                 kconf_resource = ws_k8s_data.get_kubeconfig()
    #                 if kconf_resource:
    #                     # TODO: convert this into a function
    #                     # self._ws_data_map[
    #                     #     ws_name
    #                     # ].ws_pak8_conf.k8s.kubeconfig_resource = kconf_resource
    #                     self.save_config()
    #         return self._ws_data_map[ws_name].ws_pak8_conf
    #     return None

    # def get_ws_pak8_conf_by_path(
    #     self, ws_dir_path: Path, refresh: bool = False
    # ) -> Optional[Pak8Conf]:
    #     if ws_dir_path in self._path_to_ws_data_map:
    #         if refresh:
    #             ws_data: PhiWsData = self._path_to_ws_data_map[ws_dir_path]
    #             self.refresh_ws_config(ws_data.ws_name)
    #         return self._path_to_ws_data_map[ws_dir_path].ws_pak8_conf
    #     return None

    ######################################################
    ## GCP
    ######################################################

    # def update_ws_gcp_data(
    #     self,
    #     ws_name: str,
    #     gcp_project_id: Optional[str] = None,
    #     gcp_project: Optional[GCPProjectSchema] = None,
    #     gke_cluster: Optional[GKEClusterSchema] = None,
    #     gcloud_default_creds_avl: Optional[bool] = None,
    #     gcp_svc_account: Optional[Dict[str, Any]] = None,
    #     gcp_svc_account_key: Optional[Dict[str, Any]] = None,
    # ) -> Optional[PhiGcpData]:
    #     """This function creates/updates the PhiGcpData for ws_name.
    #     When creating the PhiGcpData for the first time, the gcp_project_id must be provided.
    #     """
    #     if ws_name not in self._ws_data_map:
    #         return None
    #
    #     ws_data = self._ws_data_map[ws_name]
    #     ws_gcp_data: Optional[PhiGcpData] = ws_data.ws_gcp_data
    #     # If this is the first time we are creating the PhiGcpData for this workspace
    #     if ws_gcp_data is None:
    #         logger.debug(f"Creating PhiGcpData for {ws_name}")
    #         if gcp_project_id is None:
    #             logger.error(
    #                 "gcp_project_id must be provided when creating PhiGcpData for the first time"
    #             )
    #             return None
    #         ws_gcp_data = PhiGcpData(
    #             gcp_project_id=gcp_project_id,
    #             gcp_project=gcp_project,
    #             gke_cluster=gke_cluster,
    #             gcloud_default_creds_avl=gcloud_default_creds_avl
    #             if gcloud_default_creds_avl
    #             else False,
    #             gcp_svc_account=gcp_svc_account,
    #             gcp_svc_account_key=gcp_svc_account_key,
    #         )
    #         self._ws_data_map[ws_name].ws_gcp_data = ws_gcp_data
    #         self.save_config()
    #         return ws_gcp_data
    #
    #     logger.debug(f"Updating PhiGcpData for {ws_name}")
    #     if gcp_project_id is not None:
    #         if gcp_project_id != ws_gcp_data.gcp_project_id:
    #             print_warning(
    #                 "Updating GCP Project Id from {} to {}".format(
    #                     ws_gcp_data.gcp_project_id, gcp_project_id
    #                 )
    #             )
    #         ws_gcp_data.gcp_project_id = gcp_project_id
    #     if gcp_project is not None:
    #         ws_gcp_data.gcp_project = gcp_project
    #     if gke_cluster is not None:
    #         ws_gcp_data.gke_cluster = gke_cluster
    #     if gcloud_default_creds_avl is not None:
    #         ws_gcp_data.gcloud_default_creds_avl = gcloud_default_creds_avl
    #     if gcp_svc_account is not None:
    #         ws_gcp_data.gcp_svc_account = gcp_svc_account
    #     if gcp_svc_account_key is not None:
    #         ws_gcp_data.gcp_svc_account_key = gcp_svc_account_key
    #         ws_data.pending_actions = ws_data.pending_actions - {
    #             WorkspaceSetupActions.GCP_SVC_ACCOUNT_IS_AVL
    #         }
    #         logger.debug(f"Pending Actions: {ws_data.pending_actions}")
    #
    #     self.save_config()
    #     return ws_gcp_data
    #
    # def get_ws_gcp_data_by_name(self, ws_name: str) -> Optional[PhiGcpData]:
    #     if ws_name in self._ws_data_map:
    #         return self._ws_data_map[ws_name].ws_gcp_data
    #     return None

    # def get_gcp_project_schema_by_ws_name(
    #     self, ws_name: str
    # ) -> Optional[GCPProjectSchema]:
    #     if ws_name in self._ws_data_map:
    #         ws_gcp_data: Optional[PhiGcpData] = self._ws_data_map[ws_name].ws_gcp_data
    #         if ws_gcp_data:
    #             return ws_gcp_data.gcp_project
    #     return None

    # def get_gke_cluster_by_ws_name(self, ws_name: str) -> Optional[GKEClusterSchema]:
    #     if ws_name in self._ws_data_map:
    #         ws_gcp_data: Optional[PhiGcpData] = self._ws_data_map[ws_name].ws_gcp_data
    #         if ws_gcp_data:
    #             return ws_gcp_data.gke_cluster
    #     return None

    ######################################################
    ## ReleaseSchema Data
    ######################################################

    # def get_ws_releases_by_name(self, ws_name: str) -> Optional[WsReleases]:
    #     if ws_name in self._ws_data_map:
    #         return self._ws_data_map[ws_name].ws_releases
    #     return None
    #
    # def add_release_for_ws(self, ws_name: str, release: ReleaseSchema) -> None:
    #     if (ws_name in self._ws_data_map) and release:
    #         logger.debug(f"Adding release for ws {ws_name}")
    #         self._ws_data_map[ws_name].ws_releases.add_release(release)
    #     self.save_config()
    #
    # def get_latest_release_for_ws(self, ws_name: str) -> Optional[ReleaseSchema]:
    #     if ws_name in self._ws_data_map:
    #         return self._ws_data_map[ws_name].ws_releases.get_latest_release()
    #     return None

    ######################################################
    ## K8s functions
    ######################################################

    # def get_k8s_resources_dir_path_by_ws_name(self, ws_name: str) -> Optional[Path]:
    #     ws_dir_path: Optional[Path] = None
    #     if ws_name and ws_name in self._ws_data_map:
    #         ws_dir_path = self._ws_data_map[ws_name].ws_dir_path
    #     if ws_dir_path:
    #         return ws_dir_path.joinpath("phi").joinpath("prd")
    #     return None
    #
    # def get_ws_k8s_data_by_name(self, ws_name: str) -> Optional[PhiK8sData]:
    #     if ws_name in self._ws_data_map:
    #         return self._ws_data_map[ws_name].ws_k8s_data
    #     return None
    #
    # def update_ws_k8s_data(
    #     self,
    #     ws_name: str,
    #     kubeconfig_resource: Optional[Kubeconfig] = None,
    #     kubeconfig_path: Optional[Path] = None,
    # ) -> Optional[PhiK8sData]:
    #     """This function creates/updates the PhiK8sData for ws_name.
    #     When creating the PhiK8sData for the first time, the kubeconfig_resource must be provided.
    #     """
    #     if ws_name not in self._ws_data_map:
    #         return None
    #
    #     ws_data = self._ws_data_map[ws_name]
    #     ws_k8s_data: Optional[PhiK8sData] = ws_data.ws_k8s_data
    #     # If this is the first time we are creating the PhiK8sData for this workspace
    #     if ws_k8s_data is None:
    #         logger.debug(f"Creating PhiK8sData for {ws_name}")
    #         if kubeconfig_resource is None:
    #             logger.error(
    #                 "kubeconfig_resource must be provided when creating PhiK8sData for the first time"
    #             )
    #             return None
    #         ws_k8s_data = PhiK8sData.create_using_kubeconfig_resource(
    #             ws_name=ws_name, kconf_resource=kubeconfig_resource
    #         )
    #         self._ws_data_map[ws_name].ws_k8s_data = ws_k8s_data
    #         self.save_config()
    #         return ws_k8s_data
    #
    #     logger.debug(f"Updating PhiK8sData for {ws_name}")
    #     if kubeconfig_resource is not None:
    #         ws_k8s_data.update_kubeconfig_resource(kconf_resource=kubeconfig_resource)
    #     self.save_config()
    #     return ws_k8s_data

    ######################################################
    ## PhiConf functions
    ######################################################

    def save_config(self):
        # logger.debug(f"Saving config to {str(PHI_CONF_PATH)}")
        pickle_object_to_file(self, PHI_CONF_PATH)

    @classmethod
    def get_saved_conf(cls):
        # logger.debug(f"Reading phidata conf at {PHI_CONF_PATH}")
        if PHI_CONF_PATH.exists():
            # logger.debug(f"{PHI_CONF_PATH} exists")
            if PHI_CONF_PATH.is_file():
                return unpickle_object_from_file(
                    file_path=PHI_CONF_PATH, verify_class=cls
                )
            elif PHI_CONF_PATH.is_dir():
                logger.debug(f"{PHI_CONF_PATH} is a directory, deleting and exiting")
                delete_from_fs(PHI_CONF_PATH)
        return None

    def print_to_cli(self):
        print_heading("Phidata Config\n")
        if self.user:
            print_info(f"User: {self.user.email}")
        if self.active_ws_name:
            print_info(f"Primary workspace: {self.active_ws_name}")
        if len(self._ws_data_map) > 0:
            print_subheading("\nAvailable workspaces")
            for k, v in self._ws_data_map.items():
                print_info(f"name: {v.ws_name}")
                print_info(f"dir_path: {v.ws_dir_path}")
                print_info(f"config_file_path: {v.ws_config_file_path}")
                print_horizontal_line()



from pathlib import Path

APP_NAME = "phicli"
APP_VERSION = "0.1.0"

DEFAULT_WS_NAME = "data"

PHI_CONF_DIR: Path = Path.home().resolve().joinpath(".phi")
PHI_CONF_PATH: Path = PHI_CONF_DIR.joinpath("conf")
PHI_ACCESSTOKEN_PATH: Path = PHI_CONF_DIR.joinpath("access_token")
PHI_CREDS_PATH: Path = PHI_CONF_DIR.joinpath("creds")
# PHI_KUBE_DIR: Path = PHI_CONF_DIR.joinpath("kube")
# SYSTEM_KUBE_DIR: Path = Path.home().resolve().joinpath(".kube")
# SYSTEM_KUBECONFIG_DIR: Path = SYSTEM_KUBE_DIR.joinpath("config")

# Rest API Constants
PHI_SESSION_COOKIE_KEY: str = "__phi_session"
PHI_SIGNIN_URL_WITHOUT_PARAMS: str = "http://devphi/signin"
BACKEND_API_URL: str = "http://devphi/services/backend/api/v1/"

# Logger Names
PHI_LOGGER_NAME = "phi"
PHIDATA_LOGGER_NAME = "phidata"



# from typing import Any, Dict, List, Optional, Union, cast
#
# import httpx
#
# from pak8.conf import Pak8Conf
# from phi import exceptions, schemas
# from phi.utils.common import (
#     isinstanceany,
#     pprint_info,
#     pprint_error,
#     pprint_info,
#     pprint_network_error,
# )
# from phi.utils.log import logger
# from phi.zeus_api.client import ZeusApi, get_authenticated_client, validate_response
#
#
# def get_k8s_manifests_for_workspace(
#     id_workspace: int, pak8_conf: Pak8Conf
# ) -> Optional[List[Dict[str, Any]]]:
#
#     logger.debug("Get K8s Resources")
#     authenticated_client = get_authenticated_client()
#     if authenticated_client is None:
#         return None
#     with authenticated_client as api:
#         try:
#             r: httpx.Response = api.post(
#                 "{}/{}".format(ZeusApi.get_k8s_manifests, id_workspace),
#                 data=pak8_conf.json(),
#             )
#             validate_response(r)
#         except httpx.NetworkError as e:
#             pprint_network_error()
#             return None
#
#         # logger.debug("url: {}".format(r.url))
#         # logger.debug("status: {}".format(r.status_code))
#         # logger.debug("json: {}".format(r.json()))
#         resp_data: Union[Dict[Any, Any], List[Any]] = r.json()
#
#         # logger.debug("resp_data type: {}".format(type(resp_data)))
#         if not isinstanceany(resp_data, [dict, list]):
#             raise exceptions.ZeusK8sApiException(
#                 "Could not parse zeus K8s response. Expected `list` or `dict`, Received: {}".format(
#                     type(resp_data)
#                 )
#             )
#
#         if isinstance(resp_data, dict):
#             if (
#                 r.status_code == httpx.codes.BAD_REQUEST
#                 and resp_data.get("status", "FAIL") == "FAIL"
#             ):
#                 pprint_error(resp_data.get("message", resp_data.get("detail", "FAIL")))
#
#         if r.status_code == httpx.codes.OK:
#             return cast(List[Dict[str, Any]], resp_data)
#
#     return None
#
#
# def apply_k8s_manifests_for_workspace(id_workspace: int, pak8_conf: Pak8Conf) -> bool:
#     """Makes an API call to zeus to apply all k8s manifests and
#     returns True if the action was successful
#     """
#
#     logger.debug("Apply K8s Resources")
#     authenticated_client = get_authenticated_client()
#     if authenticated_client is None:
#         return False
#     with authenticated_client as api:
#         try:
#             r: httpx.Response = api.post(
#                 "{}/{}".format(ZeusApi.apply_k8s_manifests, id_workspace),
#                 data=pak8_conf.json(),
#             )
#             validate_response(r)
#         except httpx.NetworkError as e:
#             pprint_network_error()
#             return False
#
#         logger.debug("url: {}".format(r.url))
#         logger.debug("status: {}".format(r.status_code))
#         logger.debug("json: {}".format(r.json()))
#         resp_data: Union[Dict[Any, Any], List[Any]] = r.json()
#
#         # logger.debug("resp_data type: {}".format(type(resp_data)))
#         if not isinstanceany(resp_data, [dict, list]):
#             raise exceptions.ZeusK8sApiException(
#                 "Could not parse zeus K8s response. Expected `list` or `dict`, Received: {}".format(
#                     type(resp_data)
#                 )
#             )
#
#         if isinstance(resp_data, dict):
#             if (
#                 r.status_code == httpx.codes.BAD_REQUEST
#                 and resp_data.get("status", "FAIL") == "FAIL"
#             ):
#                 pprint_error(resp_data.get("message", resp_data.get("detail", "FAIL")))
#             elif r.status_code == httpx.codes.OK:
#                 message_log = resp_data.get("message_log", "Failed")
#                 if resp_data.get("status", "FAIL") == "SUCCESS":
#                     pprint_info(
#                         "Message: {}".format(resp_data.get("message", "Failed"))
#                     )
#                     pprint_info("Log:\n  - {}".format("\n  - ".join(message_log)))
#                     return True
#                 else:
#                     pprint_error(
#                         "Message: {}".format(resp_data.get("message", "Failed"))
#                     )
#                     pprint_info("Log:\n  - {}".format("\n  - ".join(message_log)))
#
#     return False
#
#
# def get_active_resources(
#     id_workspace: int,
#     id_project: str,
#     id_cluster: str,
#     type_filters: Optional[List[str]] = None,
#     name_filters: Optional[List[str]] = None,
# ) -> Optional[List[Any]]:
#     """Makes an API call to zeus to get all active k8s objects
#     TODO: Fix this when needed
#     """
#
#     logger.debug("Get Active K8s Objects")
#     authenticated_client = get_authenticated_client()
#     if authenticated_client is None:
#         return None
#     with authenticated_client as api:
#         try:
#             r: httpx.Response = api.post(
#                 ZeusApi.get_active_resources,
#                 params={
#                     "id_workspace": id_workspace,
#                     "id_project": id_project,
#                     "id_cluster": id_cluster,
#                 },
#                 data=schemas.K8sFilters(
#                     type_filters=type_filters,
#                     name_filters=name_filters,
#                 ).dict(),
#             )
#             validate_response(r)
#         except httpx.NetworkError as e:
#             pprint_network_error()
#             return None
#
#         logger.debug("url: {}".format(r.url))
#         logger.debug("status: {}".format(r.status_code))
#         logger.debug("json: {}".format(r.json()))
#         resp_data: Union[Dict[Any, Any], List[Any]] = r.json()
#
#         logger.debug("resp_data type: {}".format(type(resp_data)))
#         if not isinstanceany(resp_data, [dict, list]):
#             raise exceptions.ZeusK8sApiException(
#                 "Could not parse zeus K8s response. Expected `list` or `dict`, Received: {}".format(
#                     type(resp_data)
#                 )
#             )
#
#         return None
#     #     if isinstance(resp_data, dict):
#     #         if (
#     #             r.status_code == httpx.codes.BAD_REQUEST
#     #             and resp_data.get("status", "FAIL") == "FAIL"
#     #         ):
#     #             pprint_error(resp_data.get("message", resp_data.get("detail", "FAIL")))
#     #         elif r.status_code == httpx.codes.OK:
#     #             message_log = resp_data.get("message_log", "Failed")
#     #             if resp_data.get("status", "FAIL") == "SUCCESS":
#     #                 pprint_info("Message: {}".format(resp_data.get("message", "Failed")))
#     #                 pprint_info("Log:\n  - {}".format("\n  - ".join(message_log)))
#     #                 return True
#     #             else:
#     #                 pprint_error(
#     #                     "Message: {}".format(resp_data.get("message", "Failed"))
#     #                 )
#     #                 pprint_info("Log:\n  - {}".format("\n  - ".join(message_log)))
#
#     # return False



from os import environ
from pathlib import Path
from typing import Optional, Dict, List

from phidata import constants
from phidata.product import DataProduct
from phidata.workflow import Workflow
from phidata.workspace import WorkspaceConfig
from phidata.types.context import PathContext, RunContext

from phi.types.run_status import RunStatus
from phi.utils.cli_console import (
    print_error,
    print_info,
    print_heading,
    print_dict,
    print_info,
    print_subheading,
)
from phi.utils.log import logger
from phi.workspace.phi_ws_data import PhiWsData


def set_local_env(
    ws_config: WorkspaceConfig,
    path_context: PathContext,
) -> None:
    """VERY IMPORTANT TO RUN BEFORE RUNNING ANY LOCAL WORKFLOWS
    This function updates the local environment with Paths and Configuration
    Main uses:
        1. Path env variables used by DataAssets for building file_paths
        2. Cloud config variables used by AwsAssets like aws_region and profile
    """
    # 1. Path env variables used by DataAssets for building file_paths
    if path_context.scripts_dir is not None:
        environ[constants.SCRIPTS_DIR_ENV_VAR] = str(path_context.scripts_dir)
    if path_context.storage_dir is not None:
        environ[constants.STORAGE_DIR_ENV_VAR] = str(path_context.storage_dir)
    if path_context.meta_dir is not None:
        environ[constants.META_DIR_ENV_VAR] = str(path_context.meta_dir)
    if path_context.products_dir is not None:
        environ[constants.PRODUCTS_DIR_ENV_VAR] = str(path_context.products_dir)
    if path_context.notebooks_dir is not None:
        environ[constants.NOTEBOOKS_DIR_ENV_VAR] = str(path_context.notebooks_dir)
    if path_context.workspace_config_dir is not None:
        environ[constants.WORKSPACE_CONFIG_DIR_ENV_VAR] = str(
            path_context.workspace_config_dir
        )
    # 2. Cloud config variables used by AwsAssets like aws_region and profile
    if ws_config.aws_region is not None:
        environ[constants.AWS_REGION_ENV_VAR] = ws_config.aws_region
    if ws_config.aws_profile is not None:
        environ[constants.AWS_PROFILE_ENV_VAR] = ws_config.aws_profile
    if ws_config.aws_config_file is not None:
        environ[constants.AWS_CONFIG_FILE_ENV_VAR] = ws_config.aws_config_file
    if ws_config.aws_shared_credentials_file is not None:
        environ[
            constants.AWS_SHARED_CREDENTIALS_FILE_ENV_VAR
        ] = ws_config.aws_shared_credentials_file
    if ws_config.local_env:
        environ.update(ws_config.local_env)


def run_workflows_local(
    workflow_file: str,
    workflows: Dict[str, Workflow],
    run_context: RunContext,
    ws_data: PhiWsData,
    workflow_name: Optional[str] = None,
) -> bool:
    """
    Runs a workflow in the local environment

    Args:
        workflow_file: Path of the workflow file relative to the products_dir.
            This is used to build the path_context
        workflows: Dict[str, Workflow] to run
        run_context: RunContext
        ws_data: PhiWsData
        workflow_name:

    Returns:
        run_status (bool): True is the Workflow ran successfully
    """

    # Validate
    if ws_data.ws_config is None:
        print_error("WorkspaceConfig invalid")
        return False
    if ws_data.ws_dir_path is None:
        print_error("Workspace directory invalid")
        return False

    # Step 1: Build the PathContext for the workflows.
    ws_config = ws_data.ws_config
    ws_dir_path = ws_data.ws_dir_path
    scripts_dir_path = ws_dir_path.joinpath(ws_config.scripts_dir).resolve()
    storage_dir_path = ws_dir_path.joinpath(ws_config.storage_dir).resolve()
    meta_dir_path = ws_dir_path.joinpath(ws_config.meta_dir).resolve()
    products_dir_path = ws_dir_path.joinpath(ws_config.products_dir).resolve()
    notebooks_dir_path = ws_dir_path.joinpath(ws_config.notebooks_dir).resolve()
    workspace_config_dir_path = ws_dir_path.joinpath(
        ws_config.workspace_config_dir
    ).resolve()
    workflow_file_path = Path(products_dir_path).joinpath(workflow_file)
    wf_path_context: PathContext = PathContext(
        scripts_dir=scripts_dir_path,
        storage_dir=storage_dir_path,
        meta_dir=meta_dir_path,
        products_dir=products_dir_path,
        notebooks_dir=notebooks_dir_path,
        workspace_config_dir=workspace_config_dir_path,
        workflow_file=workflow_file_path,
    )
    logger.debug(f"PathContext: {wf_path_context.json(indent=4)}")

    ######################################################################
    # NOTE: VERY IMPORTANT TO RUN BEFORE RUNNING ANY LOCAL WORKFLOWS
    # Update the local environment with Paths and Configuration
    ######################################################################
    set_local_env(ws_config=ws_config, path_context=wf_path_context)

    # Step 3: Run single Workflow if workflow_name is provided
    if workflow_name is not None:
        if workflow_name not in workflows:
            print_error(
                "Could not find '{}' in {}".format(
                    workflow_name, "[{}]".format(", ".join(workflows.keys()))
                )
            )
            return False

        wf_run_status: bool = False
        workflow_to_run = workflows[workflow_name]
        _name = workflow_name or workflow_to_run.name
        print_subheading(f"\nRunning {_name}")
        # Pass down context
        workflow_to_run.run_context = run_context
        workflow_to_run.path_context = wf_path_context
        wf_run_success = workflow_to_run.run_in_local_env()

        print_subheading("\nWorkflow run status:")
        print_info("{}: {}".format(_name, "Success" if wf_run_success else "Fail"))
        print_info("")
        return wf_run_success
    # Step 4: Run all Workflows if workflow_name is None
    else:
        wf_run_status: List[RunStatus] = []
        for wf_name, wf_obj in workflows.items():
            _name = wf_name or wf_obj.name
            print_subheading(f"\nRunning {_name}")
            # Pass down context
            wf_obj.run_context = run_context
            wf_obj.path_context = wf_path_context
            run_success = wf_obj.run_in_local_env()
            wf_run_status.append(RunStatus(_name, run_success))

        print_subheading("\nWorkflow run status:")
        print_info(
            "\n".join(
                [
                    "{}: {}".format(wf.name, "Success" if wf.success else "Fail")
                    for wf in wf_run_status
                ]
            )
        )
        print_info("")
        for _run in wf_run_status:
            if not _run.success:
                return False
        return True


def run_data_products_local(
    workflow_file: str,
    data_products: Dict[str, DataProduct],
    run_context: RunContext,
    ws_data: PhiWsData,
) -> bool:
    """

    Args:
        workflow_file: Path of the workflow file relative to the products_dir.
            This is used to build the path_context
        data_products: Dict[str, DataProduct] to run
        run_context: RunContext
        ws_data: PhiWsData

    Returns:
        run_status (bool): True is the DataProduct ran successfully
    """

    # Validate
    if ws_data.ws_dir_path is None:
        print_error("Workspace directory invalid")
        return False
    if ws_data.ws_config is None:
        print_error("WorkspaceConfig invalid")
        return False

    # Step 1: Build the PathContext for the DataProducts.
    ws_config = ws_data.ws_config
    ws_dir_path = ws_data.ws_dir_path
    scripts_dir_path = ws_dir_path.joinpath(ws_config.scripts_dir).resolve()
    storage_dir_path = ws_dir_path.joinpath(ws_config.storage_dir).resolve()
    meta_dir_path = ws_dir_path.joinpath(ws_config.meta_dir).resolve()
    products_dir_path = ws_dir_path.joinpath(ws_config.products_dir).resolve()
    notebooks_dir_path = ws_dir_path.joinpath(ws_config.notebooks_dir).resolve()
    workspace_config_dir_path = ws_dir_path.joinpath(
        ws_config.workspace_config_dir
    ).resolve()
    workflow_file_path = Path(products_dir_path).joinpath(workflow_file)
    dp_path_context: PathContext = PathContext(
        scripts_dir=scripts_dir_path,
        storage_dir=storage_dir_path,
        meta_dir=meta_dir_path,
        products_dir=products_dir_path,
        notebooks_dir=notebooks_dir_path,
        workspace_config_dir=workspace_config_dir_path,
        workflow_file=workflow_file_path,
    )
    logger.debug(f"PathContext: {dp_path_context.json(indent=4)}")

    ######################################################################
    # NOTE: VERY IMPORTANT TO RUN BEFORE RUNNING ANY LOCAL WORKFLOWS
    # Update the local environment with Paths and Configuration
    ######################################################################
    set_local_env(ws_config=ws_config, path_context=dp_path_context)

    # Step 3: Run the DataProducts
    dp_run_status: List[RunStatus] = []
    for dp_name, dp_obj in data_products.items():
        _name = dp_name or dp_obj.name
        print_subheading(f"\nRunning {_name}")
        # Pass down context
        dp_obj.run_context = run_context
        dp_obj.path_context = dp_path_context
        run_success = dp_obj.run_in_local_env()
        dp_run_status.append(RunStatus(_name, run_success))

    print_subheading("DataProduct run status:")
    print_info(
        "\n".join(
            [
                "{}: {}".format(dp.name, "Success" if dp.success else "Fail")
                for dp in dp_run_status
            ]
        )
    )
    print_info("")
    for _run in dp_run_status:
        if not _run.success:
            return False
    return True



from phi.utils.enums import ExtendedEnum


class UserAuthProviderEnum(ExtendedEnum):
    TEST = "TEST"
    MANUAL = "MANUAL"
    EMAIL_PASS = "EMAIL_PASS"
    GOOGLE = "GOOGLE"
    GITHUB = "GITHUB"


class VersionControlProviderEnum(ExtendedEnum):
    GITHUB = "GITHUB"
    GITLAB = "GITLAB"
    BITBUCKET = "BITBUCKET"



from typing import Optional

from phi.conf.constants import PHI_CREDS_PATH
from phi.utils.common import pickle_object_to_file, unpickle_object_from_file
from phi.utils.log import logger


class PhiCliCreds:
    def __init__(self, session_cookie: str):
        self._session_cookie = session_cookie

    @property
    def session_cookie(self) -> str:
        return self._session_cookie


def save_session_cookie(session_cookie: str):
    # logger.debug(f"Storing {session_cookie} to {str(PHI_CREDS_PATH)}")
    creds = PhiCliCreds(session_cookie)
    pickle_object_to_file(creds, PHI_CREDS_PATH)


def read_session_cookie() -> Optional[str]:
    # logger.debug(f"Reading session_cookie from {str(PHI_CREDS_PATH)}")
    creds: PhiCliCreds = unpickle_object_from_file(PHI_CREDS_PATH)
    return creds.session_cookie



import datetime
from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from phi.enums.user_enums import UserAuthProviderEnum, VersionControlProviderEnum
from phi.utils.dttm import dttm_str_to_dttm


######################################################
# User Schemas
# These should match ~/philab/services/zeus/zeus/schemas/user_schemas.py
######################################################
# The UserSchema schema defines the properties which are returned by the zeus API
# Need to be careful that this doesn't contain any sensitive info
class UserSchema(BaseModel):
    username: str
    email: str
    user_auth_provider: UserAuthProviderEnum
    is_active: bool
    is_employee: bool = False
    is_test: bool = False
    version_control_provider: Optional[VersionControlProviderEnum] = None
    id_user: int
    scopes: Optional[List[str]] = None
    name: Optional[str] = None
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    locale: Optional[str] = None
    country_code: Optional[str] = None
    tos_accepted: bool = False
    email_verified: Optional[bool] = False
    create_ts: datetime.datetime
    last_update_ts: Optional[datetime.datetime] = None

    @classmethod
    def from_dict(cls, user_dict: Dict[str, Any]):
        return cls(
            username=user_dict.get("username", None),
            email=user_dict.get("email", None),
            user_auth_provider=UserAuthProviderEnum.from_str(
                user_dict.get("user_auth_provider", None)
            ),
            is_active=user_dict.get("is_valid", None),
            is_employee=user_dict.get("is_employee", None),
            is_test=user_dict.get("is_test", None),
            version_control_provider=VersionControlProviderEnum.from_str(
                user_dict.get("version_control_provider", None)
            ),
            id_user=user_dict.get("id_user", None),
            scopes=user_dict.get("scopes", None),
            name=user_dict.get("name", None),
            first_name=user_dict.get("first_name", None),
            last_name=user_dict.get("last_name", None),
            locale=user_dict.get("locale", None),
            country_code=user_dict.get("country_code", None),
            tos_accepted=user_dict.get("tos_accepted", None),
            email_verified=user_dict.get("email_verified", None),
            create_ts=dttm_str_to_dttm(user_dict.get("create_ts", None)),
            last_update_ts=dttm_str_to_dttm(user_dict.get("last_update_ts", None)),
        )



import datetime
from pathlib import Path
from typing import List, Optional, Set

from pydantic import BaseModel
from phidata.workspace import WorkspaceConfig

from phi.utils.common import is_empty
from phi.utils.dttm import current_datetime_utc
from phi.utils.log import logger
from phi.workspace.ws_enums import WorkspaceSetupActions
from phi.workspace.ws_exceptions import WorkspaceConfigException
from phi.workspace.ws_schemas import WorkspaceSchema


class PhiWsData(BaseModel):
    """The PhiWsData model stores data for a phidata workspace."""

    # Name of the workspace
    ws_name: str
    # WorkspaceSchema: If exists then indicates that this ws has been authenticated
    # with the backend
    ws_schema: Optional[WorkspaceSchema] = None
    # The root directory for the workspace.
    # This field indicates that this ws has been downloaded on this machine
    ws_dir_path: Optional[Path] = None
    # WorkspaceConfig for the workspace
    ws_config: Optional[WorkspaceConfig] = None
    # Path for the WorkspaceConfig file
    ws_config_file_path: Optional[Path] = None
    # A Set of WorkspaceSetupActions which this workspace must satisfy to become valid.
    # if len(ws_data.required_actions.intersection(ws_data.pending_actions)) > 0:
    # the workspace is invalid
    required_actions: Set[WorkspaceSetupActions] = {
        WorkspaceSetupActions.WS_CONFIG_IS_AVL
    }
    # A Set of WorkspaceSetupActions which this workspace currently has not
    # completed and needs to fulfill.
    # When (pending_actions is None or len(pending_actions) == 0), the workspace is valid
    pending_actions: Set[WorkspaceSetupActions] = set()
    # Timestamp of when this workspace was created on the users machine
    create_ts: datetime.datetime = current_datetime_utc()
    # Timestamp when this ws was last updated
    last_update_ts: Optional[datetime.datetime] = None

    class Config:
        arbitrary_types_allowed = True

    ######################################################
    ## WorkspaceConfig functions
    ######################################################

    def load_workspace_from_config_file(self) -> None:

        if self.ws_config_file_path is None or self.ws_dir_path is None:
            raise WorkspaceConfigException("WorkspaceConfig invalid")

        from phi.workspace.ws_loader import load_workspace

        logger.debug(f"**--> Loading WorkspaceConfig")
        ws_configs: List[WorkspaceConfig]
        ws_configs = load_workspace(self.ws_config_file_path)

        if is_empty(ws_configs):
            logger.debug(f"No WorkspaceConfig found")
            self.ws_config = None
            raise WorkspaceConfigException("No WorkspaceConfig found")

        if len(ws_configs) > 1:
            logger.warning(
                "Found {} WorkspaceConfigs, first one will be selected".format(
                    len(ws_configs)
                )
            )
        ws_config: WorkspaceConfig = ws_configs[0]
        # logger.debug("loading workspace config:\n{}".format(ws_config.args))
        self.ws_config = ws_config

    ######################################################
    ## Print functions
    ######################################################

    def print_to_cli(self):
        from rich.pretty import pprint

        pprint(self.dict())



"""Phi Workflow Cli

This is the entrypoint for the `phi wf` commands
"""

from typing import Optional

import typer

from phi.utils.cli_console import (
    print_error,
    print_info,
    print_conf_not_available_msg,
    print_active_workspace_not_available,
    print_available_workspaces,
)
from phi.utils.dttm import yesterday_datetime, dttm_to_dttm_str
from phi.utils.log import logger, set_log_level_to_debug
from phi.workflow.wf_enums import WorkflowEnv

wf_app = typer.Typer(
    name="wf",
    short_help="Commands to build and run workflows",
    help="""\b
Use `phi wf <command>` to build and run workflows.
Run `phi wf <command> --help` for more info.
    """,
    no_args_is_help=True,
    add_completion=False,
    invoke_without_command=True,
    options_metavar="\b",
    subcommand_metavar="<command>",
)


@wf_app.command(short_help="Run workflow", options_metavar="\b", no_args_is_help=True)
def run(
    wf_description: str = typer.Argument(
        ...,
        help="Workflow description. Format - Dir/File:Workflow. Where `Dir` is the path to the workflow dir relative to the products dir",
        metavar="[wf_desc]",
    ),
    env_filter: Optional[str] = typer.Option(
        "local",
        "-e",
        "--env",
        metavar="",
        help="The environment to run the workflow in. Default: local. Available Options: {}".format(
            WorkflowEnv.values_list()
        ),
    ),
    app_name: Optional[str] = typer.Option(
        None,
        "-a",
        metavar="",
        help="The App to run the workflow if using docker or k8s. Default: `devbox` for docker and `databox` for K8s.",
    ),
    run_date: Optional[str] = typer.Option(
        None, "-dt", help="Run datetime for the workflow. Default: yesterday"
    ),
    dry_run: bool = typer.Option(
        False,
        "-dr",
        "--dry-run",
        help="Perform a dry run for each task. Does not run the task.",
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """\b
    Run a workflow in the target environment.

    \b
    Examples:
    $ phi wf run dau -> Runs file: dau/dau.py, data products: all
    $ phi wf run dau:save -> Runs file: dau/dau.py, workflow: save
    $ phi wf run dau/dau2:save -> Runs file: dau/dau2.py, workflow: save
    """
    from phi.conf.phi_conf import PhiConf, PhiWsData
    from phi.workflow.wf_operator import run_workflow

    if print_debug_log:
        set_log_level_to_debug()

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return

    active_ws_data: Optional[PhiWsData] = phi_conf.get_active_ws_data(refresh=True)
    if active_ws_data is None:
        print_active_workspace_not_available()
        avl_ws = phi_conf.available_ws
        if avl_ws:
            print_available_workspaces(avl_ws)
        return

    target_env: Optional[WorkflowEnv] = None
    target_app: Optional[str] = app_name
    target_dt: str = run_date or dttm_to_dttm_str(
        yesterday_datetime(), dttm_format="%Y-%m-%d"
    )

    target_env_str = env_filter or active_ws_data.ws_config.default_env or "dev"
    try:
        target_env = WorkflowEnv.from_str(target_env_str)
    except Exception as e:
        print_error(e)
        print_error(
            f"{target_env_str} is not supported, please choose from: {WorkflowEnv.values_list()}"
        )
        return

    logger.debug("Running workflow")
    logger.debug(f"\tworkflow     : {wf_description}")
    logger.debug(f"\ttarget_env   : {target_env}")
    logger.debug(f"\ttarget_dt    : {target_dt}")
    logger.debug(f"\tdry_run      : {dry_run}")
    logger.debug(f"\ttarget_app   : {target_app}")
    run_workflow(
        wf_description=wf_description,
        ws_data=active_ws_data,
        target_env=target_env,
        target_dt=target_dt,
        dry_run=dry_run,
        target_app=target_app,
    )


@wf_app.command(short_help="Build workflow")
def build(workflow: str = typer.Argument(..., help="The workflow to build")):
    print_info(f"Building workflow {workflow}")
    print_info(f"To be implemented")


@wf_app.command(short_help="List workflows", options_metavar="\b", no_args_is_help=True)
def ls(
    wf_description: str = typer.Argument(
        ...,
        help="Workflow description. Format - Dir/File:Workflow. Where `Dir` is the path to the workflow dir relative to the products dir",
        metavar="[wf_desc]",
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    from phi.conf.phi_conf import PhiConf, PhiWsData
    from phi.workflow.wf_operator import list_workflows

    if print_debug_log:
        set_log_level_to_debug()

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return

    active_ws_data: Optional[PhiWsData] = phi_conf.get_active_ws_data(refresh=True)
    if active_ws_data is None:
        print_active_workspace_not_available()
        avl_ws = phi_conf.available_ws
        if avl_ws:
            print_available_workspaces(avl_ws)
        return

    list_workflows(
        wf_description=wf_description,
        ws_data=active_ws_data,
    )



from pathlib import Path
from typing import Optional, List

from pydantic import ValidationError
from phidata.workspace import WorkspaceConfig

from phi.utils.cli_console import print_error, print_info, print_validation_errors
from phi.utils.log import logger


def load_workspace(config_file_path: Path) -> List[WorkspaceConfig]:
    """
    Loads a workspace using a config_file_path and returns List[WorkspaceConfig]

    NOTE: When loading a workspace, relative imports or package imports dont work.
    This is a known problem in python :(
        eg: https://stackoverflow.com/questions/6323860/sibling-package-imports/50193944#50193944
    To make them work, we need to install the workspace as a python module
    But when we run `phi ws setup`, the ws module is not yet installed
    So we add workspace_root to sys.path so is treated as a package
    """

    import importlib.util
    from importlib.machinery import ModuleSpec

    logger.debug(f"--^^-- Loading workspace from: {config_file_path}")

    # Add workspace_root to sys.path so is treated as a package
    try:
        import sys

        workspace_config_dir_path = config_file_path.parent.resolve()
        workspace_root_dir_path = workspace_config_dir_path.parent
        workspace_parent_dir_path = workspace_root_dir_path.parent.resolve()
        logger.debug(f"Adding {workspace_parent_dir_path} to path")
        sys.path.insert(0, str(workspace_parent_dir_path))
    except Exception as e:
        print_error(
            "Could not add workspace parent directory to Path, this will break references to the workspace config"
        )
        logger.exception(e)

    # Read the WorkspaceConfig objects from config_file_path
    try:
        # https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
        # Create a ModuleSpec
        ws_config_module_spec: Optional[
            ModuleSpec
        ] = importlib.util.spec_from_file_location("ws_config_module", config_file_path)
        # Using the ModuleSpec create a module
        if ws_config_module_spec:
            ws_config_module = importlib.util.module_from_spec(ws_config_module_spec)
            ws_config_module_spec.loader.exec_module(ws_config_module)  # type: ignore

            # loop over all objects in module and find WorkspaceConfig objects
            ws_configs: List[WorkspaceConfig] = []
            for k, v in ws_config_module.__dict__.items():
                # logger.debug(f"Found {k}: | Type: {v.__class__.__name__}")
                if isinstance(v, WorkspaceConfig):
                    ws_configs.append(v)
            logger.debug("--^^-- Loading complete")
            return ws_configs
    except NameError as name_err:
        print_error("Variable not defined")
        raise
    except ValidationError as validation_err:
        print_error(str(validation_err))
        # print_validation_errors(validation_err.errors())
        print_info("Please fix and try again")
        exit(0)
    except (ModuleNotFoundError, Exception) as e:
        # logger.exception(e)
        raise



from collections import OrderedDict
from pathlib import Path
from typing import Optional, Dict, Tuple

from phidata.product import DataProduct
from phidata.workflow import Workflow
from phi.utils.cli_console import print_error, print_info, print_validation_errors
from phi.utils.log import logger


def get_workflow_file_and_name(
    wf_description: str, ws_dir_path: Path, products_dir: str
) -> Tuple[Optional[str], Optional[str]]:
    """
    Given a workflow description string, returns the workflow_file and workflow_name.
    The workflow_file is the path of the file starting after the products_dir.
    The workflow_file string can then be joined to the `products_dir` in the current env
    to get the absolute workflow file path
    Examples:

    `dau` -> (dau/dau.py, None)
    `dau:save` -> (dau/dau.py, save)
    `dau/dau2:save` -> (dau/dau2.py, save)

    Args:
        wf_description:
        ws_dir_path:
        products_dir:

    Returns:
        workflow_file: string path to the file containing the workflow,
                        relative to the products dir
        workflow_name: name of the workflow (if provided)
    """

    # Path to the file containing the workflow
    workflow_file: Optional[str] = None
    workflow_name: Optional[str] = None

    # Parse the workflow file and name
    _wf_path_slices = wf_description.strip().split(":")
    _wf_path: str = _wf_path_slices[0]
    _wf_py_file: str = _wf_path + ".py"
    workflow_name = _wf_path_slices[1] if len(_wf_path_slices) > 1 else None

    # Check if products dir is valid
    products_dir_path: Path = ws_dir_path.joinpath(products_dir)
    if not (products_dir_path.exists() and products_dir_path.is_dir()):
        logger.error("Invalid products dir: {}".format(products_dir))
        return None, None

    # Check if the workflow is a file or a directory
    _wf_file_check: Path = products_dir_path.joinpath(_wf_py_file)
    _wf_dir_check: Path = products_dir_path.joinpath(_wf_path)

    if _wf_file_check.exists() and _wf_file_check.is_file():
        workflow_file = _wf_py_file
    elif _wf_dir_check.exists() and _wf_dir_check.is_dir():
        _wf_file_in_dir_check = _wf_dir_check.joinpath(_wf_py_file)
        if _wf_file_in_dir_check.exists() and _wf_file_in_dir_check.is_file():
            workflow_file = "{}/{}".format(_wf_path, _wf_py_file)

    if workflow_file is None:
        # logger.error("Workflow file not found")
        return None, None

    return workflow_file, workflow_name


def get_data_products_and_workflows_from_file(
    file_path: Path,
) -> Tuple[Dict[str, DataProduct], Dict[str, Workflow]]:
    """
    Reads the DataProducts & Workflows from filepath
    Args:
        file_path:

    Returns:

    """

    import importlib.util
    from importlib.machinery import ModuleSpec
    from pydantic import ValidationError

    logger.debug(f"Reading {file_path}")

    # Read DataProduct and Workflow objects from file_path
    try:
        # https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
        # Create a ModuleSpec
        dp_module_spec: Optional[ModuleSpec] = importlib.util.spec_from_file_location(
            "dp_module", file_path
        )
        # Using the ModuleSpec create a module
        if dp_module_spec:
            dp_module = importlib.util.module_from_spec(dp_module_spec)
            dp_module_spec.loader.exec_module(dp_module)  # type: ignore

            # loop over all objects in module and find DataProduct and Workflow objects
            dp_dict: Dict[str, DataProduct] = OrderedDict()
            wf_dict: Dict[str, Workflow] = OrderedDict()
            for k, v in dp_module.__dict__.items():
                if isinstance(v, DataProduct):
                    logger.debug(f"Found {k} | Type: {v.__class__.__name__}")
                    dp_dict[k] = v
                if isinstance(v, Workflow):
                    logger.debug(f"Found {k} | Type: {v.__class__.__name__}")
                    wf_dict[k] = v
            logger.debug("--^^-- Loading complete")
            return dp_dict, wf_dict
    except NameError as name_err:
        print_error("Variable not defined")
        raise
    except ValidationError as validation_err:
        print_error(str(validation_err))
        # print_validation_errors(validation_err.errors())
        print_info("Please fix and try again")
        exit(0)
    except (ModuleNotFoundError, Exception) as e:
        # logger.exception(e)
        raise



# from typing import Optional
#
# from pak8.conf import Pak8Conf
# from phi import schemas
# from phi.cli.gcp.gke import gke_operator
# from phi.conf.phi_conf import PhiConf
# from phi.utils.common import pprint_info, pprint_error, pprint_heading
#
#
# def deploy_workspace_to_gcp(
#     ws_schema: schemas.WorkspaceSchema, config: PhiConf, refresh_pak8_conf: bool = False
# ) -> None:
#
#     pprint_heading(f"Deploying {ws_schema.name} to gcp")
#
#     # Step 1: Get the Pak8Conf for this workspace
#     ws_name: str = ws_schema.name
#     ws_pak8_conf: Optional[Pak8Conf] = config.get_ws_pak8_conf_by_name(
#         ws_name, refresh=refresh_pak8_conf
#     )
#     if ws_pak8_conf is None:
#         pprint_error(f"Unable to get config for {ws_name}")
#         return None
#
#     # Step 2: Check pending actions
#     # ws_pending_actions: Optional[
#     #     Set[enums.WorkspaceSetupActions]
#     # ] = config.get_ws_pending_actions_by_name(ws_name)
#     # if enums.WorkspaceSetupActions.GCP_SVC_ACCOUNT_IS_AVL in ws_pending_actions:
#     #     pprint_error(
#     #         "A GCP Service Account is not available for this workspace, run `phi gcp auth`"
#     #     )
#     #     return None
#
#     # Step 3: Get the GCPProjectSchema for this workspace
#     gcp_project: Optional[
#         schemas.GCPProjectSchema
#     ] = config.get_gcp_project_schema_by_ws_name(ws_name)
#     if gcp_project is None:
#         pprint_error("No GCP Project for this workspace, please run `phi gcp auth`")
#         return None
#
#     # Step 4: Get or create the GKEClusterSchema for this workspace
#     gke_cluster: Optional[
#         schemas.GKEClusterSchema
#     ] = gke_operator.get_or_create_gke_cluster(ws_schema, config, refresh_pak8_conf)
#     if gke_cluster:
#         gcloud_cmd = (
#             f"gcloud container clusters get-credentials {gke_cluster.cluster_name}"
#         )
#         pprint_info(f"To authenticate with gcloud, please run:\n\n\t{gcloud_cmd}\n")
#
#     # id_ws = ws_schema.id_workspace
#     # id_prj = gcp_project.id_project
#     # gke_cluster: Optional[schemas.GKEClusterSchema] = None
#     # if id_ws and id_prj:
#     #     pprint_status("Creating a GKE cluster if needed, this could take a while...")
#     #
#     #     # gke_cluster_release: Optional[
#     #     #     schemas.GKEClusterRelease
#     #     # ] = zeus_api.get_or_create_gke_cluster_for_gcp_project(
#     #     #     id_workspace=id_ws, id_project=id_prj, pak8_conf=ws_pak8_conf
#     #     # )
#     #     # if gke_cluster_release is not None:
#     #     #     # Note, since we're not providing the gcp_project_id arg, this would fail if
#     #     #     # ws_gcp_data is not setup.
#     #     #     gke_cluster = gke_cluster_release.gke_cluster
#     #     #     gke_create_cluster_release = gke_cluster_release.release
#     #     #     if gke_cluster is not None:
#     #     #         config.update_ws_gcp_data(ws_name=ws_name, gke_cluster=gke_cluster)
#     #     #     if gke_create_cluster_release is not None:
#     #     #         config.add_release_for_ws(
#     #     #             ws_name=ws_name, release=gke_create_cluster_release
#     #     #         )
#
#     # Step 5: Deploy all Pak8 Services
#
#     # Step 6: Validate



from httpx import Response, codes

from phi.conf.constants import APP_NAME, APP_VERSION
from phi.utils.cli_console import log_auth_error_msg, log_server_error_msg

headers = {"user-agent": f"{APP_NAME}/{APP_VERSION}"}


def response_is_invalid(r: Response) -> bool:
    """returns true if the response is invalid"""

    if r.status_code in (
        codes.UNAUTHORIZED,
        codes.BAD_REQUEST,
    ):
        log_auth_error_msg()
        return True
    if codes.is_server_error(r.status_code):
        log_server_error_msg()
        return True
    return False



# from phi.utils.enums import ExtendedEnum
#
#
# class GCPProjectStatus(ExtendedEnum):
#
#     ## Deployment States
#
#     # The project has just been created, no action has yet been taken on this
#     # The GCPProjectSchema maintains this status till it is deployed or activated
#     # using activate_gcp_project(), we should try and minimize the projects in this state
#     NEW = "NEW"
#
#     # Project is re-actived after being deactivated
#     REACTIVATED = "REACTIVATED"
#
#     # A deployment for this project has been created
#     # this is a temporary state and instances of this status in the DB imply unsuccessful deployments
#     RELEASE_CREATED = "RELEASE_CREATED"
#     # A GKEClusterSchema for this project has been created, this is also temporary state
#     GKE_CLUSTER_CREATED = "GKE_CLUSTER_CREATED"
#     # We have started to deploy applications, this is also temporary state
#     K8S_APPS_DEPLOYING = "K8S_APPS_DEPLOYING"
#     # K8s Applications are now deployed, this is also temporary state
#     K8S_APPS_DEPLOYED = "K8S_APPS_DEPLOYED"
#     # The deployed K8s Applications are now validated, meaning that the deployment was successful
#     # Majority of our GCPProjects should be in this state
#     ALL_SYSTEMS_RUNNING = "ALL_SYSTEMS_RUNNING"
#
#     # K8s Applications could not be deployed, this is an error state
#     K8S_APPS_DEPLOYMENT_ERROR = "K8S_APPS_DEPLOYMENT_ERROR"
#
#     ## Shutdown states
#
#     GKE_CLUSTER_DELETE_REQUEST = "GKE_CLUSTER_DELETE_REQUEST"
#     GKE_CLUSTER_DELETED = "GKE_CLUSTER_DELETED"
#
#     # Project is requested to be shutdown
#     DEACTIVATE_REQUEST = "DEACTIVATE_REQUEST"
#     # Project is shutdown
#     DEACTIVATED = "DEACTIVATED"
#
#     UNSPECIFIED = "UNSPECIFIED"
#
#
# class GCPOperationType(ExtendedEnum):
#     """Currently supported GCP Operations."""
#
#     # Operation types from GKE
#     # https://cloud.google.com/kubernetes-engine/docs/reference/rest/v1beta1/projects.locations.operations#Operation.Type
#     GKE_CREATE_CLUSTER = "GKE_CREATE_CLUSTER"
#     GKE_DELETE_CLUSTER = "GKE_DELETE_CLUSTER"
#     GKE_TYPE_UNSPECIFIED = "GKE_TYPE_UNSPECIFIED"
#     GKE_UPGRADE_MASTER = "GKE_UPGRADE_MASTER"
#     GKE_UPGRADE_NODES = "GKE_UPGRADE_NODES"
#     GKE_REPAIR_CLUSTER = "GKE_REPAIR_CLUSTER"
#     GKE_UPDATE_CLUSTER = "GKE_UPDATE_CLUSTER"
#     GKE_CREATE_NODE_POOL = "GKE_CREATE_NODE_POOL"
#     GKE_DELETE_NODE_POOL = "GKE_DELETE_NODE_POOL"
#     GKE_SET_NODE_POOL_MANAGEMENT = "GKE_SET_NODE_POOL_MANAGEMENT"
#     GKE_AUTO_REPAIR_NODES = "GKE_AUTO_REPAIR_NODES"
#     GKE_AUTO_UPGRADE_NODES = "GKE_AUTO_UPGRADE_NODES"
#     GKE_SET_LABELS = "GKE_SET_LABELS"
#     GKE_SET_MASTER_AUTH = "GKE_SET_MASTER_AUTH"
#     GKE_SET_NODE_POOL_SIZE = "GKE_SET_NODE_POOL_SIZE"
#     GKE_SET_NETWORK_POLICY = "GKE_SET_NETWORK_POLICY"
#     GKE_SET_MAINTENANCE_POLICY = "GKE_SET_MAINTENANCE_POLICY"



import datetime
from typing import Any, Dict, Optional

from pydantic import BaseModel

from phi.utils.cli_console import print_info, print_subheading
from phi.utils.dttm import current_datetime_utc
from phi.utils.log import logger

######################################################
# ReleaseSchema Schemas
# These should match ~/philab/services/zeus/zeus/schemas/release_schemas.py
######################################################


class ReleaseSchema(BaseModel):
    id_release: str
    is_active: Optional[bool] = None
    is_at_desired_state: Optional[bool] = None
    release_type: Optional[str] = None
    release_data: Dict[str, Any] = {}
    start_ts: datetime.datetime = current_datetime_utc()
    end_ts: Optional[datetime.datetime] = None
    at_desired_state_ts: Optional[datetime.datetime] = None

    @classmethod
    def from_dict(cls, release_dict: Dict[str, Any]):
        return cls(
            id_release=release_dict.get("id_release", None),
            is_active=release_dict.get("is_valid", None),
            is_at_desired_state=release_dict.get("is_at_desired_state", None),
            release_type=release_dict.get("release_type", None),
            release_data=release_dict.get("release_data", {}),
            start_ts=release_dict.get("start_ts", current_datetime_utc()),
            end_ts=release_dict.get("end_ts", None),
            at_desired_state_ts=release_dict.get("at_desired_state_ts", None),
        )

    def set_at_desired_state(self):
        if self.is_at_desired_state:
            logger.debug(f"Release {self.id_release} already at desired state")
        else:
            logger.debug(f"Setting release {self.id_release} at desired state")
            self.is_at_desired_state = True
            self.at_desired_state_ts = current_datetime_utc()

    def print_info(self):
        print_subheading("Release: {}".format(self.id_release))
        print_info("Active: {}".format(self.is_active))
        print_info("At desired state: {}".format(self.is_at_desired_state))
        print_info("Release Type: {}".format(self.release_type))


#
#
# class ReleaseData(BaseModel):
#     # id_workspace of the WorkspaceSchema
#     ws__id_workspace: Optional[int]
#     # id_project of the GCPProjectSchema
#     ws_gcp_project__id_project: Optional[str]
#     # gcp project id
#     ws_gcp_project__gcp_project_id: Optional[str]
#     # id_cluster of the GKEClusterSchema
#     ws_gke_cluster__id_cluster: Optional[str]
#     # GKEClusterSchema name
#     ws_gke_cluster__cluster_name: Optional[str]
#
#     # other deployment datapoints
#     create_gke_cluster__id_operation: Optional[str]
#     delete_gke_cluster__id_operation: Optional[str]
#
#     class Config:
#         use_enum_values = True
#
#     @classmethod
#     def from_dict(cls, release_data_dict: Dict[str, Any]):
#         return cls(
#             ws__id_workspace=release_data_dict.get("ws__id_workspace", None),
#             ws_gcp_project__id_project=release_data_dict.get(
#                 "ws_gcp_project__id_project", None
#             ),
#             ws_gcp_project__gcp_project_id=release_data_dict.get(
#                 "ws_gcp_project__gcp_project_id", None
#             ),
#             ws_gke_cluster__id_cluster=release_data_dict.get(
#                 "ws_gke_cluster__id_cluster", None
#             ),
#             ws_gke_cluster__cluster_name=release_data_dict.get(
#                 "ws_gke_cluster__cluster_name", None
#             ),
#             create_gke_cluster__id_operation=release_data_dict.get(
#                 "create_gke_cluster__id_operation", None
#             ),
#             delete_gke_cluster__id_operation=release_data_dict.get(
#                 "delete_gke_cluster__id_operation", None
#             ),
#         )


# class GKEClusterRelease(BaseModel):
#     gke_cluster: GKEClusterSchema
#     release: ReleaseSchema
#
#     @classmethod
#     def from_dict(cls, gke_cluster_release_dict: Dict[str, Any]):
#         return cls(
#             gke_cluster=GKEClusterSchema.from_dict(
#                 gke_cluster_release_dict.get("gke_cluster", {})
#             ),
#             release=ReleaseSchema.from_dict(
#                 gke_cluster_release_dict.get("release", {})
#             ),
#         )



import datetime
from typing import List, Optional

from pydantic import BaseModel

from phi.release.release_enums import ReleaseType
from phi.release.release_schemas import ReleaseSchema
from phi.utils.dttm import current_datetime_utc
from phi.utils.log import logger


class WsReleases(BaseModel):
    # This data will grow and should be stored using a sqlite db on the users machine
    releases: List[ReleaseSchema] = []
    last_update_ts: datetime.datetime = current_datetime_utc()

    def add_release(self, rls: ReleaseSchema):
        self.releases.append(rls)
        self.last_update_ts = current_datetime_utc()

    def get_latest_release(self) -> Optional[ReleaseSchema]:
        if len(self.releases) > 0:
            return self.releases[-1]
        return None

    def get_releases_by_type(
        self, rel_type: ReleaseType, active_only: bool = True
    ) -> Optional[List[ReleaseSchema]]:
        """Returns active releases of type rel_type"""
        rls = [rel for rel in self.releases if rel.release_type == rel_type.value]
        if rls:
            if active_only:
                return [r for r in rls if r.is_active]
            return rls
        return None

    def set_active_releases_as_complete(self, rel_type: ReleaseType) -> None:
        """Sets active releases of rel_type as complete"""

        logger.debug(f"Marking existing {rel_type.value} releases as complete")

        active_releases: Optional[List[ReleaseSchema]] = self.get_releases_by_type(
            rel_type
        )
        if active_releases is None or len(active_releases) == 0:
            logger.debug(f"No active {rel_type.value} releases found")
            return

        for rls in active_releases:
            rls.is_active = False
            rls.end_ts = current_datetime_utc()
            # rls.last_update_ts = current_datetime_utc()
            logger.debug(
                f"Release {rls.id_release} ({rls.release_type}) marked as complete"
            )


#
# def secho_releases(ws_releases: WsReleases, last_n_releases: int = 3) -> None:
#
#     pprint_subheading("\nPast Releases")
#     releases_by_type: Dict[
#         str, List[ReleaseSchema]
#     ] = ws_releases.releases_by_type
#     for rel_type, rel_list in releases_by_type.items():
#         pprint_subheading(f"\n{rel_type}:")
#         for i in range(min(len(rel_list), last_n_releases)):
#             rel_list[i].secho_status()
